# -*- coding: utf-8 -*-
"""validate"""


def validate_smearing(parameters, smearing=None):
    """Validate smearing parameters and update the parameters input node accordingly.
    :param parameters: the Dict node that will be used in the inputs
    :param smearing: a tuple of a string and float corresponding to type of smearing and the degauss value
    :raises ValueError: if the input is invalid
    """
    if not any(smearing):
        return

    valid_smearing_types = {
        "gaussian": ["gaussian", "gauss"],
        "methfessel-paxton": ["methfessel-paxton", "m-p", "mp"],
        "cold": ["marzari-vanderbilt", "cold", "m-v", "mv"],
        "fermi-dirac": ["fermi-dirac", "f-d", "fd"],
    }

    for _, options in valid_smearing_types.items():
        if smearing[0] in options:
            break
    else:
        raise ValueError(
            f"the smearing type \"{smearing[0]}\" is invalid, choose from {', '.join(list(valid_smearing_types.keys()))}"
        )

    if not isinstance(smearing[1], float):
        raise ValueError("the smearing value should be a float")

    parameters["SYSTEM"]["occupations"] = "smearing"
    parameters["SYSTEM"]["smearing"] = smearing[0]
    parameters["SYSTEM"]["degauss"] = smearing[1]



# -*- coding: utf-8 -*-
"""
WorkChain calculate the delta for certain pseudopotential
"""
from aiida import orm
from aiida.plugins import DataFactory
from plumpy import ToContext

from aiida_sssp_workflow.calculations.calculate_delta import delta_analyze
from aiida_sssp_workflow.workflows.evaluate._eos import _EquationOfStateWorkChain

from . import _BaseEvaluateWorkChain

UpfData = DataFactory("pseudo.upf")


class DeltaWorkChain(_BaseEvaluateWorkChain):
    """WorkChain calculate the bands for certain pseudopotential"""

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.expose_inputs(_EquationOfStateWorkChain, namespace='eos')
        spec.input('element', valid_type=orm.Str,
                    help='element')
        spec.input('configuration', valid_type=orm.Str,
                    help='Configuration name of structure, determine with AE reference data to use.')
        # TODO: validate structure in _EOS conform with element and configuration.

        spec.outline(
            cls.run_eos,
            cls.inspect_eos,
            cls.finalize,
        )
        spec.expose_outputs(_EquationOfStateWorkChain, namespace='eos',
                    namespace_options={'help': f'volume_energy and birch_murnaghan_fit result from EOS.'})

        spec.output('output_parameters', required=True,
                    help='The output of delta factor and other measures to describe the accuracy of EOS compare '
                        ' with the AE equation of state.')

        # yapf: enable

    def _get_inputs(self):
        inputs = self.exposed_inputs(_EquationOfStateWorkChain, namespace="eos")

        return inputs

    def run_eos(self):
        """run eos workchain"""
        inputs = self._get_inputs()

        running = self.submit(_EquationOfStateWorkChain, **inputs)
        self.report(f"launching _EquationOfStateWorkChain<{running.pk}>")

        return ToContext(eos=running)

    def inspect_eos(self):
        """Inspect the results of _EquationOfStateWorkChain"""
        workchain = self.ctx.eos

        if not workchain.is_finished_ok:
            self.logger.warning(
                f"_EquationOfStateWorkChain failed with exit status {workchain.exit_status}"
            )

        self.ctx.ecutwfc = workchain.inputs.pw.parameters["SYSTEM"]["ecutwfc"]
        self.ctx.ecutrho = workchain.inputs.pw.parameters["SYSTEM"]["ecutrho"]

        self.out_many(
            self.exposed_outputs(
                workchain,
                _EquationOfStateWorkChain,
                namespace="eos",
            )
        )

    def finalize(self):
        """result"""
        output_bmf = self.outputs["eos"].get("output_birch_murnaghan_fit")

        V0 = orm.Float(output_bmf["volume0"])
        natoms = output_bmf["num_of_atoms"]

        inputs = {
            "element": self.inputs.element,
            "configuration": self.inputs.configuration,
            "V0": V0,
            "B0": orm.Float(output_bmf["bulk_modulus0"]),
            "B1": orm.Float(output_bmf["bulk_deriv0"]),
            "natoms": orm.Int(natoms),
        }

        # set ecutwfc and ecutrho
        self.out("ecutwfc", orm.Int(self.ctx.ecutwfc).store())
        self.out("ecutrho", orm.Int(self.ctx.ecutrho).store())

        self.out(f"output_parameters", delta_analyze(**inputs))



# -*- coding: utf-8 -*-
# pylint: disable=wrong-import-position,wildcard-import
"""Module for the command line interface."""
import click
import click_completion
from aiida.cmdline.params import options as options_core
from aiida.cmdline.params import types

# Activate the completion of parameter types provided by the click_completion package
click_completion.init()


@click.group(
    "aiida-sssp-workflow", context_settings={"help_option_names": ["-h", "--help"]}
)
@options_core.PROFILE(type=types.ProfileParamType(load_profile=True))
def cmd_root(profile):  # pylint: disable=unused-argument
    """CLI for the `aiida-sssp-workflow` plugin."""


from aiida_sssp_workflow.cli.tools import dump_output



# -*- coding: utf-8 -*-
"""
Convergence test on bands of a given pseudopotential
"""

from aiida import orm
from aiida.engine import calcfunction
from aiida.plugins import DataFactory

from aiida_sssp_workflow.calculations.calculate_bands_distance import get_bands_distance
from aiida_sssp_workflow.utils import NONMETAL_ELEMENTS, update_dict
from aiida_sssp_workflow.workflows.convergence._base import _BaseConvergenceWorkChain
from aiida_sssp_workflow.workflows.evaluate._bands import BandsWorkChain

UpfData = DataFactory("pseudo.upf")


@calcfunction
def helper_bands_distence_difference(
    band_structure_a: orm.BandsData,
    band_parameters_a: orm.Dict,
    band_structure_b: orm.BandsData,
    band_parameters_b: orm.Dict,
    smearing: orm.Float,
    fermi_shift: orm.Float,
    do_smearing: orm.Bool,
):
    """doc"""
    # `get_bands_distance` require the less electrons results pass
    # to inputs of label 'a', do swap a and b if not.
    num_electrons_a = band_parameters_a["number_of_electrons"]
    num_electrons_b = band_parameters_b["number_of_electrons"]

    if num_electrons_a > num_electrons_b:
        band_parameters_a, band_parameters_b = band_parameters_b, band_parameters_a
        band_structure_a, band_structure_b = band_structure_b, band_structure_a

    res = get_bands_distance(
        band_structure_a,
        band_structure_b,
        band_parameters_a,
        band_parameters_b,
        smearing.value,
        fermi_shift.value,
        do_smearing.value,
    )
    eta = res.get("eta_c", None)
    shift = res.get("shift_c", None)
    max_diff = res.get("max_diff_c", None)

    return orm.Dict(
        dict={
            "eta_c": eta * 1000,
            "shift_c": shift * 1000,
            "max_diff_c": max_diff * 1000,
            "bands_unit": "meV",  # unit mev with value * 1000
        }
    )


class ConvergenceBandsWorkChain(_BaseConvergenceWorkChain):
    """WorkChain to converge test on cohisive energy of input structure"""

    # pylint: disable=too-many-instance-attributes

    _RY_TO_EV = 13.6056980659

    _PROPERTY_NAME = "bands"
    _EVALUATE_WORKCHAIN = BandsWorkChain
    _MEASURE_OUT_PROPERTY = "eta_c"

    def init_setup(self):
        super().init_setup()
        self.ctx.extra_pw_parameters = {}

    def setup_code_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        super().setup_code_parameters_from_protocol()

        # parse protocol
        protocol = self.ctx.protocol
        self.ctx.degauss = self._DEGAUSS = protocol["degauss"]
        self._OCCUPATIONS = protocol["occupations"]
        self._SMEARING = protocol["smearing"]
        self._CONV_THR = protocol["electron_conv_thr"]
        self.ctx.kpoints_distance_scf = protocol["kpoints_distance_scf"]
        self.ctx.kpoints_distance_bands = protocol["kpoints_distance_bands"]

        # Set context parameters
        self.ctx.pw_parameters = super()._get_pw_base_parameters(
            self._DEGAUSS, self._OCCUPATIONS, self._SMEARING, self._CONV_THR
        )

        self.ctx.fermi_shift = protocol["fermi_shift"]
        self.ctx.init_nbands_factor = protocol["init_nbands_factor"]
        self.ctx.is_metal = self.ctx.element not in NONMETAL_ELEMENTS

        self.logger.info(f"The parameters for convergence is: {self.ctx.pw_parameters}")

    def _get_inputs(self, ecutwfc, ecutrho):
        """
        get inputs for the evaluation CohesiveWorkChain by provide ecutwfc and ecutrho,
        all other parameters are fixed for the following steps
        """
        parameters = {
            "SYSTEM": {
                "ecutwfc": ecutwfc,
                "ecutrho": ecutrho,
            },
        }

        parameters = update_dict(parameters, self.ctx.pw_parameters)

        parameters_bands = update_dict(parameters, {})
        parameters_bands["SYSTEM"].pop("nbnd", None)
        parameters_bands["CONTROL"].pop("tstress", None)

        inputs = {
            "structure": self.ctx.structure,
            "scf": {
                "pw": {
                    "code": self.inputs.code,
                    "pseudos": self.ctx.pseudos,
                    "parameters": orm.Dict(dict=parameters),
                    "metadata": {
                        "options": self.ctx.options,
                    },
                    "parallelization": orm.Dict(dict=self.ctx.parallelization),
                },
                "kpoints_distance": orm.Float(self.ctx.kpoints_distance_scf),
            },
            "bands": {
                "pw": {
                    "code": self.inputs.code,
                    "pseudos": self.ctx.pseudos,
                    "parameters": orm.Dict(dict=parameters_bands),
                    "metadata": {
                        "options": self.ctx.options,
                    },
                    "parallelization": orm.Dict(dict=self.ctx.parallelization),
                },
            },
            "kpoints_distance_bands": orm.Float(self.ctx.kpoints_distance_bands),
            "init_nbands_factor": orm.Float(self.ctx.init_nbands_factor),
            "fermi_shift": orm.Float(self.ctx.fermi_shift),
            "run_bands_structure": orm.Bool(
                False
            ),  # for convergence with no band structure evaluate
            "clean_workchain": self.inputs.clean_workchain,
        }

        return inputs

    def helper_compare_result_extract_fun(self, sample_node, reference_node, **kwargs):
        """implement"""
        sample_band_parameters = sample_node.outputs["bands"].band_parameters
        reference_band_parameters = reference_node.outputs["bands"].band_parameters

        sample_band_structure = sample_node.outputs["bands"].band_structure
        reference_band_structure = reference_node.outputs["bands"].band_structure

        # Always process smearing to find fermi level even for non-metal elements.
        res = helper_bands_distence_difference(
            sample_band_structure,
            sample_band_parameters,
            reference_band_structure,
            reference_band_parameters,
            smearing=orm.Float(self.ctx.degauss * self._RY_TO_EV),
            fermi_shift=orm.Float(self.ctx.fermi_shift),
            do_smearing=orm.Bool(True),
        ).get_dict()

        return res

    def get_result_metadata(self):
        return {
            "fermi_shift": self.ctx.fermi_shift,
            "bands_unit": "meV",
        }



# -*- coding: utf-8 -*-
"""
All in one verification workchain
"""
# pylint: disable=cyclic-import

from typing import Optional

from aiida import orm
from aiida.engine import ToContext, if_
from aiida.engine.processes.exit_code import ExitCode
from aiida.engine.processes.functions import calcfunction
from aiida.plugins import DataFactory, WorkflowFactory

from aiida_sssp_workflow.workflows import SelfCleanWorkChain
from aiida_sssp_workflow.workflows.common import operate_calcjobs
from aiida_sssp_workflow.workflows.convergence import _BaseConvergenceWorkChain
from aiida_sssp_workflow.workflows.convergence.caching import (
    _CachingConvergenceWorkChain,
)
from aiida_sssp_workflow.workflows.measure import _BaseMeasureWorkChain

UpfData = DataFactory("pseudo.upf")


@calcfunction
def parse_pseudo_info(pseudo: UpfData):
    """parse the pseudo info as a Dict"""
    from pseudo_parser.upf_parser import parse

    try:
        info = parse(pseudo.get_content())
    except ValueError:
        return ExitCode(100, "cannot parse the info of pseudopotential.")

    return orm.Dict(dict=info)


DEFAULT_CONVERGENCE_PROPERTIES_LIST = [
    "convergence.cohesive_energy",
    "convergence.phonon_frequencies",
    "convergence.pressure",
    "convergence.delta",
    "convergence.bands",
]

DEFAULT_ACCURACY_PROPERTIES_LIST = [
    "accuracy.delta",
    "accuracy.bands",
]

DEFAULT_PROPERTIES_LIST = (
    DEFAULT_ACCURACY_PROPERTIES_LIST + DEFAULT_CONVERGENCE_PROPERTIES_LIST
)


class VerificationWorkChain(SelfCleanWorkChain):
    """The verification workflow to run all test for the given pseudopotential"""

    # This two class attributes will control whether a WF flow is
    # run and results write to outputs ports.
    _VALID_CONGENCENCE_WF = [
        "convergence.cohesive_energy",
        "convergence.phonon_frequencies",
        "convergence.pressure",
        "convergence.delta",
        "convergence.bands",
    ]
    _VALID_ACCURACY_WF = [
        "accuracy.delta",
        "accuracy.bands",
    ]

    @classmethod
    def define(cls, spec):
        super().define(spec)
        # yapf: disable
        spec.expose_inputs(_BaseMeasureWorkChain, namespace='accuracy',
                    exclude=['code', 'pseudo', 'options', 'parallelization', 'clean_workchain'])
        spec.expose_inputs(_BaseConvergenceWorkChain, namespace='convergence',
                    exclude=['code', 'pseudo', 'options', 'parallelization', 'clean_workchain'])
        spec.input('pw_code', valid_type=orm.Code,
                    help='The `pw.x` code use for the `PwCalculation`.')
        spec.input('ph_code', valid_type=orm.Code,
                    help='The `ph.x` code use for the `PhCalculation`.')
        spec.input('pseudo', valid_type=UpfData, required=True,
                    help='Pseudopotential to be verified')
        spec.input('label', valid_type=orm.Str, required=False,
                    help='label store for display as extra attribut.')
        spec.input('properties_list', valid_type=orm.List,
                    default=lambda: orm.List(list=DEFAULT_PROPERTIES_LIST),
                    help='The preperties will be calculated, passed as a list.')
        spec.input('options', valid_type=orm.Dict, required=False,
                    help='Optional `options`')
        spec.input('parallelization', valid_type=orm.Dict, required=False,
                    help='Parallelization options')

        spec.outline(
            cls.setup_code_resource_options,
            cls.parse_pseudo,
            cls.init_setup,
            if_(cls.is_verify_accuracy)(
                cls.run_accuracy,
                cls.inspect_accuracy,
            ),
            if_(cls.is_verify_convergence)(
                if_(cls.is_caching)(
                    cls.run_caching,
                    cls.inspect_caching,
                ),
                cls.run_convergence,
                cls.inspect_convergence,
            ),
        )
        spec.output('pseudo_info', valid_type=orm.Dict, required=True,
            help='pseudopotential info')
        for wfname in cls._VALID_ACCURACY_WF:
            spec.output_namespace(wfname, dynamic=True,
                help=f'results of {wfname} calculation.')
        for wfname in cls._VALID_CONGENCENCE_WF:
            spec.output_namespace(wfname, dynamic=True,
                help=f'results of {wfname} calculation.')

        spec.exit_code(401, 'ERROR_CACHING_ON_BUT_FAILED',
            message='The caching is triggered but failed.')
        spec.exit_code(811, 'WARNING_NOT_ALL_SUB_WORKFLOW_OK',
            message='The sub-workflows {processes} is not finished ok.')
        # yapf: enable

    def setup_code_resource_options(self):
        """
        setup resource options and parallelization for `PwCalculation` from inputs
        """
        if "options" in self.inputs:
            self.ctx.options = self.inputs.options.get_dict()
        else:
            from aiida_sssp_workflow.utils import get_default_options

            self.ctx.options = get_default_options(
                with_mpi=True,
            )

        if "parallelization" in self.inputs:
            self.ctx.parallelization = self.inputs.parallelization.get_dict()
        else:
            self.ctx.parallelization = {}

    @staticmethod
    def _label_from_pseudo_info(pseudo_info) -> str:
        """derive a label string from pseudo_info dict"""
        element = pseudo_info["element"]
        pp_type = pseudo_info["pp_type"]
        z_valence = pseudo_info["z_valence"]

        return f"{element}.{pp_type}.z_{z_valence}"

    def parse_pseudo(self):
        """parse pseudo"""
        pseudo_info = parse_pseudo_info(self.inputs.pseudo)
        self.ctx.pseudo_info = pseudo_info.get_dict()
        self.node.set_extra_many(
            self.ctx.pseudo_info
        )  # set the extra attributes for the node

        self.out("pseudo_info", pseudo_info)

    def init_setup(self):
        """prepare inputs for all verification process"""

        if "label" in self.inputs:
            label = self.inputs.label.value
        else:
            label = self._label_from_pseudo_info(self.ctx.pseudo_info)

        self.node.set_extra("label", label)

        # Properties list
        valid_list = self._VALID_ACCURACY_WF + self._VALID_CONGENCENCE_WF
        self.ctx.properties_list = [
            p for p in self.inputs.properties_list.get_list() if p in valid_list
        ]

        # Accuracy workflow: bands measure and delta measure workflows inputs setting
        accurary_inputs = self.exposed_inputs(
            _BaseMeasureWorkChain, namespace="accuracy"
        )
        accurary_inputs["pseudo"] = self.inputs.pseudo
        accurary_inputs["code"] = self.inputs.pw_code
        accurary_inputs["options"] = self.inputs.options
        accurary_inputs["parallelization"] = self.inputs.parallelization

        accurary_inputs["clean_workchain"] = self.inputs.clean_workchain

        self.ctx.accuracy_inputs = {
            "delta": accurary_inputs.copy(),
            "bands": accurary_inputs.copy(),
        }

        # Convergence inputs setting, the properties of convergence test are:
        # 1. cohesive energy
        # 2. phonon frequencies
        # 3. pressue
        # 4. delta
        # 5. bands distance
        convergence_inputs = self.exposed_inputs(
            _BaseConvergenceWorkChain, namespace="convergence"
        )
        convergence_inputs["code"] = self.inputs.pw_code
        convergence_inputs["pseudo"] = self.inputs.pseudo
        convergence_inputs["options"] = self.inputs.options
        convergence_inputs["parallelization"] = self.inputs.parallelization

        convergence_inputs["clean_workchain"] = self.inputs.clean_workchain

        # Here, the shallow copy can be used since the type of convergence_inputs
        # is AttributesDict.
        # The deepcopy can't be used, since it will create new data node.
        inputs_phonon_frequencies = convergence_inputs.copy()
        inputs_phonon_frequencies.pop("code", None)
        inputs_phonon_frequencies["pw_code"] = self.inputs.pw_code
        inputs_phonon_frequencies["ph_code"] = self.inputs.ph_code

        self.ctx.convergence_inputs = {
            "cohesive_energy": convergence_inputs.copy(),
            "phonon_frequencies": inputs_phonon_frequencies,
            "pressure": convergence_inputs.copy(),
            "delta": convergence_inputs.copy(),
            "bands": convergence_inputs.copy(),
        }

        self.ctx.caching_inputs = convergence_inputs.copy()
        self.ctx.caching_inputs["clean_workchain"] = orm.Bool(
            False
        )  # shouldn't clean until last, default of _caching but do it here explicitly

        # to collect workchains in a dict
        self.ctx.workchains = {}

        # For store the finished_ok workflow
        self.ctx.finished_ok_wf = {}

    def is_verify_accuracy(self):
        """
        Whether to run accuracy (delta measure, bands distance} workflow.
        """
        for p in self.ctx.properties_list:
            if "accuracy" in p:
                return True

        return False

    def run_accuracy(self):
        """Run delta measure sub-workflow"""
        for property in DEFAULT_ACCURACY_PROPERTIES_LIST:
            property_name = property.split(".")[1]
            if property in self.ctx.properties_list:
                AccuracyWorkflow = WorkflowFactory(f"sssp_workflow.{property}")

                running = self.submit(
                    AccuracyWorkflow, **self.ctx["accuracy_inputs"][property_name]
                )
                self.report(
                    f"Submit {property_name} accuracy workchain pk={running.pk}"
                )

                self.to_context(_=running)
                self.ctx.workchains[f"{property}"] = running

    def inspect_accuracy(self):
        """Inspect delta measure results"""
        return self._report_and_results(wname_list=self._VALID_ACCURACY_WF)

    def is_verify_convergence(self):
        """Whether to run convergence test workflows"""
        if "_caching" in self.ctx.properties_list:
            # for only run caching workflow
            return True

        for p in self.ctx.properties_list:
            if "convergence" in p:
                return True

        return False

    def is_caching(self):
        """run caching when more than one convergence test"""
        # If the aiida config set pw caching off, then not caching any
        from aiida.manage.caching import get_use_cache

        identifier = "aiida.calculations:quantumespresso.pw"
        if not get_use_cache(identifier=identifier):
            return False
        else:
            return True

    def run_caching(self):
        """run pressure verification for caching"""
        ##
        # Pressure as caching
        ##
        running = self.submit(_CachingConvergenceWorkChain, **self.ctx.caching_inputs)
        self.report(
            f"The caching is triggered, submit and run caching "
            f"workchain pk={running.pk} for following convergence test."
            ""
        )

        return ToContext(verify_caching=running)

    def inspect_caching(self):
        """Simply check whether caching run finished okay."""
        workchain = self.ctx.verify_caching

        if not workchain.is_finished_ok:
            return self.exit_codes.ERROR_CACHING_ON_BUT_FAILED

    def run_convergence(self):
        """
        running all verification workflows
        """
        for property in DEFAULT_CONVERGENCE_PROPERTIES_LIST:
            property_name = property.split(".")[1]
            if property in self.ctx.properties_list:
                ConvergenceWorkflow = WorkflowFactory(f"sssp_workflow.{property}")

                running = self.submit(
                    ConvergenceWorkflow, **self.ctx["convergence_inputs"][property_name]
                )
                self.report(
                    f"Submit {property_name} convergence workchain pk={running.pk}"
                )

                self.to_context(_=running)
                self.ctx.workchains[f"{property}"] = running

    def inspect_convergence(self):
        """
        inspect the convergence result

        the list set the avaliable convergence workchain that will be inspected
        """
        return self._report_and_results(wname_list=self._VALID_CONGENCENCE_WF)

    def _report_and_results(self, wname_list):
        """result to respective output namespace"""

        not_finished_ok_wf = {}
        for wname, workchain in self.ctx.workchains.items():
            if wname in wname_list:
                # dump all output as it is to verification workflow output
                self.ctx.finished_ok_wf[wname] = workchain.pk
                for label in workchain.outputs:
                    # output node and namespace -> verification workflow outputs
                    self.out(f"{wname}.{label}", workchain.outputs[label])

                if not workchain.is_finished_ok:
                    self.logger.warning(
                        f"The sub-workflow {wname} pk={workchain.pk} not finished ok."
                    )
                    not_finished_ok_wf[wname] = workchain.pk

        if not_finished_ok_wf:
            return self.exit_codes.WARNING_NOT_ALL_SUB_WORKFLOW_OK.format(
                processes=not_finished_ok_wf
            )

    def on_terminated(self):
        super().on_terminated()

        if self.inputs.clean_workchain.value is False:
            self.report(f"{type(self)}: remote folders will not be cleaned")
            return

        def _invalid_cache(node: orm.CalcJobNode) -> Optional[int]:
            """This is different from invalid_cache of `common.py`.
            It is stringent that even for non-cached node it will be invalided from caching
            so it can not be used for further caching. Only applied below for `_caching`
            workflow at the end of verification."""

            if "_aiida_hash" in node.extras:
                # It is implemented in aiida 2.0.0, by setting the is_valid_cache.
                # set the it to disable the caching to precisely control extras.
                # here in order that the correct node is cleaned and caching controlled
                # I only invalid_caching if this node is cached from other node, otherwise
                # that node (should be the node from `_caching` workflow) will not be invalid
                # caching.
                # This ensure that if the calcjob is identically running, it will still be used for
                # further calculation.
                node.delete_extra("_aiida_hash")
                return node.pk
            else:
                return None

        # For calcjobs in _caching, to prevent it from being used by second run after
        # remote work_dir cleaned. I invalid it from caching if it is being cleaned.
        invalid_calcs = operate_calcjobs(
            self.ctx.verify_caching, operator=_invalid_cache, all_same_nodes=True
        )

        if invalid_calcs:
            self.report(
                f"Invalid cache of `_caching` (even nonmenon) workflow's calcjob node: {' '.join(map(str, invalid_calcs))}"
            )



from aiida.engine import WorkChain
from aiida.orm import Bool, CalcJobNode

from aiida_sssp_workflow.workflows.common import (
    clean_workdir,
    invalid_cache,
    operate_calcjobs,
)


class SelfCleanWorkChain(WorkChain):
    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input(
            "clean_workchain",
            valid_type=Bool,
            default=lambda: Bool(True),
            help="If `True`, work directories of all called non-cached calculations will be cleaned"
            " at the end of execution, and cached calculations will invalid from cache.",
        )

    def on_terminated(self):
        """Clean the working directories of all child calculations if `clean_worchain=True` in the inputs.
        If clean, important to also invalid calcjob nodes for caching. Since otherwise there will be chance
        that not the caching node but cleaned node used for caching in bands and phonon that will lead to the
        parent_folder empty issue.
        In big verification workflow, only _caching workflow is not process clean step at last but purge remote
        folder after the whole verfication workflow finished.
        hardcode the invalid_caching.
        """
        super().on_terminated()

        if self.inputs.clean_workchain.value is False:
            self.report(f"{type(self)}: remote folders will not be cleaned")
            return

        cleaned_calcs = operate_calcjobs(
            self.node, operator=clean_workdir, all_same_nodes=False
        )
        cache_invalid_calcs = operate_calcjobs(
            self.node, operator=invalid_cache, all_same_nodes=False
        )

        if cleaned_calcs:
            self.report(
                f"cleaned remote folders of calculations: {' '.join(map(str, cleaned_calcs))}"
            )
        if cache_invalid_calcs:
            self.report(
                f"Invalid cache of cached calculations: {' '.join(map(str, cache_invalid_calcs))}"
            )



# -*- coding: utf-8 -*-
"""
Convergence test on phonon frequencies of a given pseudopotential
"""

from aiida import orm
from aiida.engine import calcfunction
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import update_dict
from aiida_sssp_workflow.workflows.convergence._base import _BaseConvergenceWorkChain
from aiida_sssp_workflow.workflows.evaluate._phonon_frequencies import (
    PhononFrequenciesWorkChain,
)

UpfData = DataFactory("pseudo.upf")


@calcfunction
def helper_phonon_frequencies_difference(
    element: orm.Str, input_parameters: orm.Dict, ref_parameters: orm.Dict
) -> orm.Dict:
    """
    The phonon frequencies are calculated at BZ boundary qpoint (1/2, 1/2, 1/2).
    The difference between the test cutoff and reference cutoff are compared.

    For some elements, we have neglected the first n frequencies in the summation above,
    because the frequencies are negative and/or with strong oscillations as
    function of the cutoff for all the considered pseudos).
    We have neglected the first 4 frequencies for H and I, 12 for N and Cl,
    6 for O and ??SiF4 (which replaces F)??.
    """
    import numpy as np

    input_frequencies = input_parameters["dynamical_matrix_0"]["frequencies"]
    ref_frequencies = ref_parameters["dynamical_matrix_0"]["frequencies"]

    # set strat_idx the idx of frequencies start to count
    element = element.value
    if element == 'N' or element == 'Cl':
        start_idx = 12
    elif element == 'H' or element == 'I':
        start_idx = 4
    elif element == 'O':
        start_idx = 6
    else:
        start_idx = 0

    input_frequencies = input_frequencies[start_idx:]
    ref_frequencies = ref_frequencies[start_idx:]

    # calculate the diff
    diffs = np.array(input_frequencies) - np.array(ref_frequencies)
    weights = np.array(ref_frequencies)

    omega_max = np.amax(input_frequencies)

    absolute_diff = np.mean(diffs)
    absolute_max_diff = np.amax(diffs)

    relative_diff = np.sqrt(np.mean((diffs / weights) ** 2))
    relative_max_diff = np.amax(diffs / weights)

    return orm.Dict(
        dict={
            "omega_max": omega_max,
            "relative_diff": relative_diff,
            "relative_max_diff": relative_max_diff,
            "absolute_diff": absolute_diff,
            "absolute_max_diff": absolute_max_diff,
            "absolute_unit": "cm-1",
            "relative_unit": "%",
        }
    )


class ConvergencePhononFrequenciesWorkChain(_BaseConvergenceWorkChain):
    """WorkChain to converge test on cohisive energy of input structure"""

    # pylint: disable=too-many-instance-attributes

    _PROPERTY_NAME = "phonon_frequencies"
    _EVALUATE_WORKCHAIN = PhononFrequenciesWorkChain
    _MEASURE_OUT_PROPERTY = "relative_diff"

    @classmethod
    def define(cls, spec):
        super().define(spec)
        # yapf: disable
        del spec.inputs['code'] # For this input code port set in base work chain need to be unset.
        spec.input("pw_code", valid_type=orm.Code,
            help="The `pw.x` code use for the `PwCalculation`.")
        spec.input("ph_code",valid_type=orm.Code,
            help="The `ph.x` code  use for the `PhCalculation`.",
        )
        # yapf: enable

    def init_setup(self):
        super().init_setup()
        self.ctx.extra_ph_parameters = {}
        self.ctx.extra_pw_parameters = {}

    def extra_setup_for_rare_earth_element(self):
        super().extra_setup_for_rare_earth_element()

        extra_ph_parameters = {
            "INPUTPH": {
                "diagonalization": "cg",
            }
        }
        self.ctx.extra_ph_parameters = update_dict(
            self.ctx.extra_ph_parameters, extra_ph_parameters
        )

    def setup_code_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        super().setup_code_parameters_from_protocol()

        protocol = self.ctx.protocol
        # PW
        self._DEGAUSS = protocol["degauss"]
        self._OCCUPATIONS = protocol["occupations"]
        self._SMEARING = protocol["smearing"]
        self._CONV_THR = protocol["electron_conv_thr"]
        self._KDISTANCE = protocol["kpoints_distance"]

        # PH
        self._QPOINTS_LIST = protocol["qpoints_list"]
        self._PH_EPSILON = protocol["epsilon"]
        self._PH_TR2_PH = protocol["tr2_ph"]

        self.ctx.qpoints_list = self._QPOINTS_LIST

        self.ctx.pw_parameters = super()._get_pw_base_parameters(
            self._DEGAUSS, self._OCCUPATIONS, self._SMEARING, self._CONV_THR
        )

        self.ctx.ph_parameters = {
            "INPUTPH": {
                "tr2_ph": self._PH_TR2_PH,
                "epsil": self._PH_EPSILON,
            }
        }

        self.ctx.ph_parameters = update_dict(
            self.ctx.ph_parameters, self.ctx.extra_ph_parameters
        )
        self.ctx.kpoints_distance = self._KDISTANCE

        self.logger.info(f"The pw parameters for convergence is: {self.ctx.pw_parameters}")
        self.logger.info(f"The ph parameters for convergence is: {self.ctx.ph_parameters}")

    def _get_inputs(self, ecutwfc, ecutrho):
        """
        get inputs for the evaluation CohesiveWorkChain by provide ecutwfc and ecutrho,
        all other parameters are fixed for the following steps
        """
        pw_parameters = {
            "SYSTEM": {
                "ecutwfc": ecutwfc,
                "ecutrho": ecutrho,
            },
        }
        pw_parameters = update_dict(pw_parameters, self.ctx.pw_parameters)

        qpoints = orm.KpointsData()
        qpoints.set_cell_from_structure(self.ctx.structure)
        qpoints.set_kpoints(self.ctx.qpoints_list)

        # convert parallelization to CMDLINE for PH
        # since ph calculation now doesn't support parallelization
        cmdline_list = []
        for key, value in self.ctx.parallelization.items():
            cmdline_list.append(f"-{str(key)}")
            cmdline_list.append(str(value))

        # Sinec PH calculation always runs more time then the correspoding pw calculation
        # set the walltime to 4 times as set in option.
        ph_options = update_dict(self.ctx.options, {})
        pw_max_walltime = self.ctx.options.get("max_wallclock_seconds", None)
        if pw_max_walltime:
            ph_options["max_wallclock_seconds"] = pw_max_walltime * 4

        inputs = {
            "scf": {
                "metadata": {"call_link_label": "SCF"},
                "pw": {
                    "structure": self.ctx.structure,
                    "code": self.inputs.pw_code,
                    "pseudos": self.ctx.pseudos,
                    "parameters": orm.Dict(dict=pw_parameters),
                    "metadata": {
                        "options": self.ctx.options,
                    },
                    "parallelization": orm.Dict(dict=self.ctx.parallelization),
                },
                "kpoints_distance": orm.Float(self.ctx.kpoints_distance),
            },
            "phonon": {
                "metadata": {"call_link_label": "PH"},
                "ph": {
                    "code": self.inputs.ph_code,
                    "qpoints": qpoints,
                    "parameters": orm.Dict(dict=self.ctx.ph_parameters),
                    "metadata": {
                        "options": ph_options,
                    },
                    "settings": orm.Dict(dict={"CMDLINE": cmdline_list}),
                },
            },
            "clean_workchain": self.inputs.clean_workchain,
        }

        return inputs

    def get_result_metadata(self):
        return {
            "absolute_unit": "cm-1",
            "relative_unit": "%",
        }

    def helper_compare_result_extract_fun(self, sample_node, reference_node, **kwargs):
        """extract"""
        sample_output = sample_node.outputs.output_parameters
        reference_output = reference_node.outputs.output_parameters

        return helper_phonon_frequencies_difference(
            orm.Str(self.ctx.element), sample_output, reference_output
        ).get_dict()



# -*- coding: utf-8 -*-
"""
Convergence test on pressure of a given pseudopotential
"""
from aiida import orm
from aiida.engine import calcfunction
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import RARE_EARTH_ELEMENTS, update_dict
from aiida_sssp_workflow.workflows.convergence._base import _BaseConvergenceWorkChain
from aiida_sssp_workflow.workflows.evaluate._eos import _EquationOfStateWorkChain
from aiida_sssp_workflow.workflows.evaluate._pressure import PressureWorkChain

UpfData = DataFactory("pseudo.upf")


def helper_get_volume_from_pressure_birch_murnaghan(P, V0, B0, B1):
    """
    Knowing the pressure P and the Birch-Murnaghan equation of state
    parameters, gets the volume the closest to V0 (relatively) that is
    such that P_BirchMurnaghan(V)=P

    retrun unit is (%)

    !! The unit of P and B0 must be compatible. We use eV/angs^3 here.
    Therefore convert P from GPa to eV/angs^3
    """
    import numpy as np

    # convert P from GPa to eV/angs^3
    P = P / 160.21766208

    # coefficients of the polynomial in x=(V0/V)^(1/3) (aside from the
    # constant multiplicative factor 3B0/2)
    polynomial = [
        3.0 / 4.0 * (B1 - 4.0),
        0,
        1.0 - 3.0 / 2.0 * (B1 - 4.0),
        0,
        3.0 / 4.0 * (B1 - 4.0) - 1.0,
        0,
        0,
        0,
        0,
        -2 * P / (3.0 * B0),
    ]
    V = min(
        [
            V0 / (x.real**3)
            for x in np.roots(polynomial)
            if abs(x.imag) < 1e-8 * abs(x.real)
        ],
        key=lambda V: abs(V - V0) / float(V0),
    )

    return abs(V - V0) / V0 * 100


@calcfunction
def helper_pressure_difference(
    input_parameters: orm.Dict,
    ref_parameters: orm.Dict,
    V0: orm.Float,
    B0: orm.Float,
    B1: orm.Float,
) -> orm.Dict:
    """
    The unit of output pressure and absolute diff is GPascal
    therefore the B0 unit should also be GPa, otherwise the results are wrong.
    """
    res_pressure = input_parameters["hydrostatic_stress"]
    ref_pressure = ref_parameters["hydrostatic_stress"]
    absolute_diff = abs(res_pressure - ref_pressure)
    relative_diff = helper_get_volume_from_pressure_birch_murnaghan(
        absolute_diff, V0.value, B0.value, B1.value
    )

    return orm.Dict(
        dict={
            "pressure": res_pressure,
            "relative_diff": relative_diff,
            "absolute_diff": absolute_diff,
            "absolute_unit": "GPascal",
            "relative_unit": "%",
        }
    )


class ConvergencePressureWorkChain(_BaseConvergenceWorkChain):
    """WorkChain to converge test on pressure of input structure"""

    # pylint: disable=too-many-instance-attributes

    _PROPERTY_NAME = "pressure"
    _EVALUATE_WORKCHAIN = PressureWorkChain
    _MEASURE_OUT_PROPERTY = "relative_diff"

    def init_setup(self):
        super().init_setup()
        self.ctx.pw_parameters = {}

    def setup_code_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        super().setup_code_parameters_from_protocol()

        protocol = self.ctx.protocol
        self._DEGAUSS = protocol["degauss"]
        self._OCCUPATIONS = protocol["occupations"]
        self._SMEARING = protocol["smearing"]
        self._CONV_THR = protocol["electron_conv_thr"]
        self._KDISTANCE = protocol["kpoints_distance"]

        self._EOS_SCALE_COUNT = protocol["scale_count"]
        self._EOS_SCALE_INCREMENT = protocol["scale_increment"]

        self.ctx.pw_parameters = super()._get_pw_base_parameters(
            self._DEGAUSS, self._OCCUPATIONS, self._SMEARING, self._CONV_THR
        )

        self.ctx.kpoints_distance = self._KDISTANCE

        self.logger.info(f"The pw parameters for convergence is: {self.ctx.pw_parameters}")

    def _get_inputs(self, ecutwfc, ecutrho):
        """
        get inputs for the evaluation CohesiveWorkChain by provide ecutwfc and ecutrho,
        all other parameters are fixed for the following steps
        """
        parameters = {
            "SYSTEM": {
                "ecutwfc": ecutwfc,
                "ecutrho": ecutrho,
            },
        }
        parameters = update_dict(parameters, self.ctx.pw_parameters)

        inputs = {
            "pw": {
                "code": self.inputs.code,
                "structure": self.ctx.structure,
                "pseudos": self.ctx.pseudos,
                "parameters": orm.Dict(dict=parameters),
                "metadata": {
                    "call_link_label": "pressure_SCF",
                    "options": self.ctx.options,
                },
                "parallelization": orm.Dict(dict=self.ctx.parallelization),
            },
            "kpoints_distance": orm.Float(self.ctx.kpoints_distance),
            "clean_workchain": self.inputs.clean_workchain,
        }

        return inputs

    def run_reference(self):
        """
        run on reference calculation
        """
        super().run_reference()

        # For pressure convergence workflow, the birch murnagen fitting result is used to
        # calculating the pressure. There is an extra workflow (run at ecutwfc of reference point)
        # for it which need to be run before the following step.

        # This workflow is shared with delta factor workchain for birch murnagan fitting.
        ecutwfc = self.ctx.reference_ecutwfc
        ecutrho = ecutwfc * self.ctx.dual
        parameters = {
            "SYSTEM": {
                "ecutwfc": round(ecutwfc),
                "ecutrho": round(ecutrho),
            },
        }
        parameters = update_dict(parameters, self.ctx.pw_parameters)
        parameters["CONTROL"].pop("tstress", None)

        # sparse kpoints and tetrahedra occupation in EOS reference calculation
        if self.ctx.element in RARE_EARTH_ELEMENTS:
            self.ctx.kpoints_distance = self._KDISTANCE + 0.05
            parameters["SYSTEM"].pop("smearing", None)
            parameters["SYSTEM"].pop("degauss", None)
            parameters["SYSTEM"]["occupations"] = "tetrahedra"


        inputs = {
            "metadata": {"call_link_label": "pressure_ref_EOS"},
            "structure": self.ctx.structure,
            "kpoints_distance": orm.Float(self.ctx.kpoints_distance),
            "scale_count": orm.Int(self._EOS_SCALE_COUNT),
            "scale_increment": orm.Float(self._EOS_SCALE_INCREMENT),
            "pw": {
                "code": self.inputs.code,
                "pseudos": self.ctx.pseudos,
                "parameters": orm.Dict(dict=parameters),
                "metadata": {
                    "options": self.ctx.options
                },
                "parallelization": orm.Dict(dict=self.ctx.parallelization),
            },
        }

        running = self.submit(_EquationOfStateWorkChain, **inputs)
        self.report(f"launching _EquationOfStateWorkChain<{running.pk}>")

        self.to_context(extra_reference=running)

    def inspect_reference(self):
        super().inspect_reference()

        workchain = self.ctx.extra_reference
        if not workchain.is_finished_ok:
            self.logger.warning(
                f"{workchain.process_label} pk={workchain.pk} for extra reference of "
                "pressure convergence run is failed with exit_code={workchain.exit_status}."
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED.format(
                label="extra_reference"
            )

        extra_reference = self.ctx.extra_reference
        extra_reference_parameters = extra_reference.outputs.output_birch_murnaghan_fit

        V0 = extra_reference_parameters["volume0"]
        B0 = extra_reference_parameters["bulk_modulus0"]  # The unit is eV/angstrom^3
        B1 = extra_reference_parameters["bulk_deriv0"]

        self.ctx.extra_parameters = {
            "V0": orm.Float(V0),
            "B0": orm.Float(B0),
            "B1": orm.Float(B1),
        }

    def get_result_metadata(self):
        return {
            "absolute_unit": "GPascal",
            "relative_unit": "%",
        }

    def helper_compare_result_extract_fun(self, sample_node, reference_node, **kwargs):
        """implement"""
        sample_output = sample_node.outputs.output_parameters
        reference_output = reference_node.outputs.output_parameters

        extra_parameters = kwargs["extra_parameters"]
        res = helper_pressure_difference(
            sample_output, reference_output, **extra_parameters
        ).get_dict()

        return res



# -*- coding: utf-8 -*-
"""
A calcfunctian create_isolate_atom
Create the structure of isolate atom
"""
from aiida import orm
from aiida.engine import append_, calcfunction
from aiida.plugins import DataFactory, WorkflowFactory

from . import _BaseEvaluateWorkChain

PwBaseWorkflow = WorkflowFactory("quantumespresso.pw.base")
UpfData = DataFactory("pseudo.upf")


@calcfunction
def create_isolate_atom(
    element: orm.Str, length=lambda: orm.Float(12.0)
) -> orm.StructureData:
    """
    create a cubic cell with length and the element kind. The atom is locate at the origin point.
    """
    from ase import Atoms

    L = length.value  # pylint: disable=invalid-name
    atom = Atoms(
        symbols=element.value,
        pbc=[True, True, True],
        cell=[L, L, L],
        positions=[
            (
                L / 2,
                L / 2,
                L / 2,
            )
        ],
    )
    structure = orm.StructureData(ase=atom)
    return structure


class CohesiveEnergyWorkChain(_BaseEvaluateWorkChain):
    """WorkChain to calculate cohisive energy of input structure"""

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.input('structure', valid_type=orm.StructureData,
                    help='Ground state structure which the verification perform')
        spec.input_namespace('pseudos', valid_type=UpfData, dynamic=True,
                    help='A mapping of `UpfData` nodes onto the kind name to which they should apply.')
        spec.input('atom_parameters', valid_type=orm.Dict,
                    help='parameters for pwscf of atom calculation for each element in structure.')
        spec.input('vacuum_length', valid_type=orm.Float,
                    help='The length of cubic cell in isolate atom calculation.')
        spec.expose_inputs(PwBaseWorkflow, namespace="bulk", exclude=["pw.structure", "pw.pseudos"])
        spec.expose_inputs(PwBaseWorkflow, namespace="atom", exclude=["pw.structure", "pw.pseudos"])

        spec.outline(
            cls.validate_structure,
            cls.run_energy,
            cls.inspect_energy,
            cls.finalize,
        )

        spec.output('output_parameters', valid_type=orm.Dict, required=True,
                    help='The output parameters include cohesive energy of the structure.')
        spec.exit_code(211, 'ERROR_SUB_PROCESS_FAILED_ATOM_ENERGY',
                    message='PwBaseWorkChain of atom energy evaluation failed.')
        spec.exit_code(212, 'ERROR_SUB_PROCESS_FAILED_BULK_ENERGY',
                    message='PwBaseWorkChain of bulk structure energy evaluation failed with exit status.')
        # yapf: enable

    def validate_structure(self):
        """Create isolate atom and validate structure"""
        # create isolate atom structure
        formula_list = self.inputs.structure.get_ase().get_chemical_symbols()
        elements = list(dict.fromkeys(formula_list))
        dict_element_and_count = {
            element: formula_list.count(element) for element in formula_list
        }

        # assert len(self.inputs.pseudos) == len(dict_element_and_count)
        dict_element_and_structure = {}
        for element in elements:
            atom_structure = create_isolate_atom(
                orm.Str(element), self.inputs.vacuum_length
            )
            dict_element_and_structure[element] = atom_structure

        self.ctx.pseudos = self.inputs.pseudos
        self.ctx.d_element_structure = dict_element_and_structure
        self.ctx.d_element_count = dict_element_and_count

    @staticmethod
    def _get_pseudo(element, pseudos):
        """
        get the pseudo by element from input pseudos dict
        the tricky is for the element name with number like in mag structure
        the pseudo get from the first found.
        """
        try:
            pseudo = pseudos[element]
        except KeyError:
            key = f"{element}1"
            pseudo = pseudos[key]

        return pseudo

    def run_energy(self):
        """set the inputs and submit atom/bulk energy evaluation parallel"""
        bulk_inputs = self.exposed_inputs(PwBaseWorkflow, namespace="bulk")
        bulk_inputs["pw"]["structure"] = self.inputs.structure
        bulk_inputs["pw"]["pseudos"] = self.inputs.pseudos

        running_bulk_energy = self.submit(PwBaseWorkflow, **bulk_inputs)
        self.report(
            f"Submit SCF calculation of bulk {self.inputs.structure.get_description()}"
        )
        self.to_context(workchain_bulk_energy=running_bulk_energy)

        for element, structure in self.ctx.d_element_structure.items():

            atom_inputs = self.exposed_inputs(PwBaseWorkflow, namespace="atom")
            atom_inputs["pw"]["structure"] = structure
            atom_inputs["pw"]["pseudos"] = {
                element: self._get_pseudo(element, self.inputs.pseudos),
            }
            atom_inputs["pw"]["parameters"] = orm.Dict(
                dict=self.inputs.atom_parameters[element]
            )

            running_atom_energy = self.submit(PwBaseWorkflow, **atom_inputs)
            self.logger.info(f"Submit atomic SCF of {element}.")
            self.to_context(workchain_atom_children=append_(running_atom_energy))

    def inspect_energy(self):
        """inspect the result of energy calculation."""

        workchain_bulk_energy = self.ctx["workchain_bulk_energy"]
        if not workchain_bulk_energy.is_finished_ok:
            self.logger.warning(
                f"PwBaseWorkChain of bulk energy evaluation failed"
                f" with exit status {workchain_bulk_energy.exit_status}"
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED_BULK_ENERGY

        self.ctx.bulk_energy = workchain_bulk_energy.outputs.output_parameters["energy"]
        calc_time = workchain_bulk_energy.outputs.output_parameters["wall_time_seconds"]

        self.ctx.ecutwfc = workchain_bulk_energy.inputs.pw.parameters["SYSTEM"][
            "ecutwfc"
        ]
        self.ctx.ecutrho = workchain_bulk_energy.inputs.pw.parameters["SYSTEM"][
            "ecutrho"
        ]

        element_energy = {}
        for child in self.ctx.workchain_atom_children:
            element = child.inputs.pw.structure.get_kind_names()[0]
            if not child.is_finished_ok:
                self.logger.warning(
                    f"PwBaseWorkChain of element={element} atom energy evaluation failed"
                    f" with exit status {child.exit_status}"
                )
                return self.exit_codes.ERROR_SUB_PROCESS_FAILED_ATOM_ENERGY

            output_parameters = child.outputs.output_parameters

            atom_free_energy = output_parameters["energy"]
            atom_smearing_energy = output_parameters["energy_smearing"]
            atom_energy = atom_free_energy - atom_smearing_energy
            element_energy[element] = atom_energy
            calc_time += output_parameters["wall_time_seconds"]

        self.ctx.calc_time = calc_time
        self.ctx.element_energy = element_energy

    def finalize(self):
        num_of_atoms = sum(self.ctx.d_element_count.values())
        cohesive_energy = self.ctx.bulk_energy
        element_energy = {}  # dict to be output for every element isolate energy
        for element, energy in self.ctx.element_energy.items():
            element_energy[f"isolate_atom_energy_{element}"] = energy
            cohesive_energy -= energy * self.ctx.d_element_count[element]

        cohesive_energy_per_atom = cohesive_energy / num_of_atoms

        parameters_dict = {
            "cohesive_energy": cohesive_energy,
            "cohesive_energy_per_atom": cohesive_energy_per_atom,
            "bulk_energy": self.ctx.bulk_energy,
            "structure_formula": self.inputs.structure.get_formula(),
            "energy_unit": "eV",
            "energy_per_atom_unit": "eV/atom",
            "total_calc_time": self.ctx.calc_time,
            "time_unit": "s",
        }
        parameters_dict.update(element_energy)
        output_parameters = orm.Dict(dict=parameters_dict)

        self.out("output_parameters", output_parameters.store())
        self.out("ecutwfc", orm.Int(self.ctx.ecutwfc).store())
        self.out("ecutrho", orm.Int(self.ctx.ecutrho).store())



# -*- coding: utf-8 -*-
"""
Base legacy work chain
"""
from abc import ABCMeta, abstractmethod

from aiida import orm
from aiida.engine import append_, if_
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import (
    HIGH_DUAL_ELEMENTS,
    MAGNETIC_ELEMENTS,
    RARE_EARTH_ELEMENTS,
    convergence_analysis,
    get_magnetic_inputs,
    get_protocol,
    get_standard_structure,
    reset_pseudos_for_magnetic,
    update_dict,
)
from aiida_sssp_workflow.workflows import SelfCleanWorkChain
from aiida_sssp_workflow.workflows.common import (
    get_extra_parameters_for_lanthanides,
    get_pseudo_element_and_type,
    get_pseudo_N,
)

UpfData = DataFactory('pseudo.upf')


class abstract_attribute(object):
    """lazy variable check: https://stackoverflow.com/a/32536493"""
    def __get__(self, obj, type):
        for cls in type.__mro__:
            for name, value in cls.__dict__.items():
                if value is self:
                    this_obj = obj if obj else type
                    raise NotImplementedError(
                        f'{this_obj!r} does not have the attribute {name!r} (abstract from class {cls.__name__!r})'
                    )

        raise NotImplementedError(
            '%s does not set the abstract attribute <unknown>', type.__name__)


class _BaseConvergenceWorkChain(SelfCleanWorkChain):
    """Base legacy workchain"""
    # pylint: disable=too-many-instance-attributes
    __metaclass__ = ABCMeta

    _PROPERTY_NAME = abstract_attribute()   # used to get convergence protocol
    _EVALUATE_WORKCHAIN = abstract_attribute()
    _MEASURE_OUT_PROPERTY = abstract_attribute()

    # Default set to True, override it in subclass to turn it off
    _RUN_WFC_TEST = True
    _RUN_RHO_TEST = True

    _NBANDS_FACTOR_FOR_REN = 1.5

    @classmethod
    def define(cls, spec):
        super().define(spec)
        # yapf: disable
        spec.input('code', valid_type=orm.Code,
                    help='The `pw.x` code use for the `PwCalculation`.')
        spec.input('pseudo', valid_type=UpfData, required=True,
                    help='Pseudopotential to be verified')
        spec.input('protocol', valid_type=orm.Str, required=True,
                    help='The calculation protocol to use for the workchain.')
        spec.input('cutoff_control', valid_type=orm.Str, required=True,
                    help='The cutoff control list to use for the workchain.')
        spec.input('criteria', valid_type=orm.Str, required=True,
                    help='Criteria for convergence measurement to give recommend cutoff pair.')
        spec.input('preset_ecutwfc', valid_type=orm.Int, required=False,
                    help='Preset wavefunction cutoff will be used and skip wavefunction test.')
        spec.input('options', valid_type=orm.Dict, required=False,
                    help='Optional `options`.')
        spec.input('parallelization', valid_type=orm.Dict, required=False,
                    help='Parallelization options')

        spec.outline(
            cls.init_setup,
            if_(cls.is_magnetic_element)(
                cls.extra_setup_for_magnetic_element,
            ),
            if_(cls.is_rare_earth_element)(
                cls.extra_setup_for_rare_earth_element,
            ),
            cls.setup_code_parameters_from_protocol,
            cls.setup_criteria_parameters_from_protocol,
            cls.setup_code_resource_options,
            cls.run_reference,
            cls.inspect_reference,
            if_(cls._is_run_wfc_convergence_test)(
                cls.run_wfc_convergence_test,
                cls.inspect_wfc_convergence_test,
            ),
            if_(cls._is_run_rho_convergence_test)(
                cls.run_rho_convergence_test,
                cls.inspect_rho_convergence_test,
            ),
            cls.finalize,
        )

        spec.output('output_parameters_wfc_test', valid_type=orm.Dict, required=False,
                    help='The output parameters include results of all wfc test calculations.')
        spec.output('output_parameters_rho_test', valid_type=orm.Dict, required=False,
                    help='The output parameters include results of all rho test calculations.')
        spec.output('output_parameters', valid_type=orm.Dict, required=False,
                    help='The output parameters of two stage convergence test.')

        spec.exit_code(401, 'ERROR_REFERENCE_CALCULATION_FAILED',
            message='The reference calculation failed.')
        spec.exit_code(402, 'ERROR_SUB_PROCESS_FAILED',
            message='The sub process for `{label}` did not finish successfully.')
        # yapy: enable

    @abstractmethod
    def helper_compare_result_extract_fun(self, sample_node, reference_node, **kwargs) -> dict:
        """
        Must be implemented for specific convergence workflow to extrac the result
        Expected to return a dict of result.

        Get the node of sample and reference as input. Extract the parameters for
        properties extract helper function.
        """

    @abstractmethod
    def get_result_metadata(self):
        """
        define a dict of which is the metadata of the results, e.g. the unit of the properties
        return a list type.

        This dict will be merged into the final `output_parameters` results.
        For different convergence workflow, you may want to add different metadata
        into output.

        for example:

        return {
            'absolute_unit': 'meV/atom',
            'relative_unit': '%',
        }
        """

    @abstractmethod
    def _get_inputs(self, ecutwfc: int, ecutrho: int) -> dict:
        """generate inputs for the evaluate workflow
        Must compatible with inputs format of every evaluate workflow.
        This is used in actual submit of reference and convergence tests.

        Since the ecutwfc and ecutrho has less point set to be a float with decimals.
        Therefore, always pass round to nearst int inputs for ecutwfc and ecutrho.
        It also bring the advantage that the inputs to final calcjob always the same
        for these two inputs and caching will properly triggered.
        """
        # TODO: there can be a validation of inputs and the evaluate workflow inputs ports

    def _is_run_wfc_convergence_test(self):
        """If running wavefunction convergence test
        default True, override class attribute `_RUN_WFC_TEST`
        in subclass to supress running it

        If 'preset_ecutwfc' is set in inputs will use that value and
        skip the wavefunction cutoff test.
        """
        if 'preset_ecutwfc' in self.inputs:
            self.ctx.wfc_cutoff = self.inputs.preset_ecutwfc.value
            assert self.ctx.wfc_cutoff < self.ctx.reference_ecutwfc

            self.ctx.output_parameters['wavefunction_cutoff'] = self.ctx.wfc_cutoff

            return False
        else:
            return self._RUN_WFC_TEST

    def _is_run_rho_convergence_test(self):
        """If running charge density convergence test
        default True, override class attribute `_RUN_RHO_TEST` in
        subclass to supress running it

        Also if self.ctx.dual_scan_list is empty means it is not set
        in control.yml protocol file. Most likely run the 200vs300_ref_check
        protocol.
        """
        return self._RUN_RHO_TEST and (self.ctx.dual_scan_list)

    def init_setup(self):
        """
        This step contains all preparation before actaul setup, e.g. set
        the context of element, base_structure, base pw_parameters and pseudos.
        """
        # init output_parameters to store output
        self.ctx.output_parameters = {}

        cutoff_control = get_protocol(category='control', name=self.inputs.cutoff_control.value)
        self.ctx.ecutwfc_list = self._ECUTWFC_LIST = cutoff_control['wfc_scan']

        # use the last cutoff as reference
        # IMPORTANT to convert to float since the value should have
        # the same tye so the caching will correctly activated.
        self.ctx.reference_ecutwfc = self._ECUTWFC_LIST[-1]

        self.ctx.extra_pw_parameters = {}
        self.ctx.element, self.ctx.pseudo_type = get_pseudo_element_and_type(self.inputs.pseudo)

        # set the ecutrho according to the type of pseudopotential
        # dual 4 for NC and 10 for all other type of PP.
        if self.ctx.pseudo_type == 'nc':
            self.ctx.dual = 4.0
            self.ctx.dual_scan_list = cutoff_control['nc_dual_scan']
        else:
            if self.ctx.element  in HIGH_DUAL_ELEMENTS:
                self.ctx.dual = 18.0
                self.ctx.dual_scan_list = cutoff_control['nonnc_high_dual_scan']
            else:
                # the initial dual set to 8 for most elements
                self.ctx.dual = 8.0

                # For the non-NC pseudos we should be careful that high charge density cutoff
                # is needed.
                # We recommond to set the scan wide range from to find the best
                # charge density cutoff.
                self.ctx.dual_scan_list = cutoff_control['nonnc_dual_scan']

        self.ctx.pseudos = {self.ctx.element: self.inputs.pseudo}

        # Please check README for what and why we use configuration set 'convergence'
        # for convergence verification.
        self.ctx.structure = get_standard_structure(self.ctx.element, prop='convergence')

    def is_magnetic_element(self):
        """Check if the element is magnetic"""
        return self.ctx.element in MAGNETIC_ELEMENTS

    def extra_setup_for_magnetic_element(self):
        """
        Extra setup for magnetic element, set starting magnetization
        and reset pseudos to correspont elements name.
        """
        self.ctx.structure, magnetic_extra_parameters = get_magnetic_inputs(self.ctx.structure)
        self.ctx.extra_pw_parameters = update_dict(self.ctx.extra_pw_parameters, magnetic_extra_parameters)

        # override pseudos setting
        # required for O, Mn, Cr where the kind names varies for sites
        self.ctx.pseudos = reset_pseudos_for_magnetic(self.inputs.pseudo, self.ctx.structure)

    def is_rare_earth_element(self):
        """Check if the element is rare earth"""
        return self.ctx.element in RARE_EARTH_ELEMENTS

    def extra_setup_for_rare_earth_element(self):
        """
        Extra setup for rare-earth element same as magnetic elements

        We use nitrdes configuration for the convergence verification of rare-earth elements.
        Otherwise it is hard to get converged in scf calculation.
        """
        nbnd_factor = self._NBANDS_FACTOR_FOR_REN
        pseudo_N = get_pseudo_N()
        self.ctx.pseudos['N'] = pseudo_N
        pseudo_RE = self.inputs.pseudo
        nbnd = nbnd_factor * (pseudo_N.z_valence + pseudo_RE.z_valence)
        self.ctx.extra_pw_parameters = get_extra_parameters_for_lanthanides(self.ctx.element, nbnd)

    def setup_code_parameters_from_protocol(self):
        """unzip and parse protocol parameters to context"""
        protocol = get_protocol(category='converge', name=self.inputs.protocol.value)
        self.ctx.protocol = {
            **protocol['base'],
            **protocol.get(self._PROPERTY_NAME, dict()),    # if _PROPERTY_NAME not set, simply use base
        }

    def _get_pw_base_parameters(self, degauss, occupations, smearing, conv_thr):
        """Return base pw parameters dict for all convengence bulk workflow
        Unchanged dict for caching purpose

        TODO: move this method out of class
        """
        parameters = {
            'SYSTEM': {
                'degauss': degauss,
                'occupations': occupations,
                'smearing': smearing,
            },
            'ELECTRONS': {
                'conv_thr': conv_thr,
            },
            'CONTROL': {
                # 'calculation': 'scf', NOT EXPLICITLY SET. `scf` is default if not override.
                'tstress': True,    # for pressue to use _caching node directly.
            },
        }

        # update with extra pw params, for magnetic and lanthenides
        if self.ctx.extra_pw_parameters:
            parameters = update_dict(parameters, self.ctx.extra_pw_parameters)

        return parameters

    def setup_criteria_parameters_from_protocol(self):
        """Input validation"""
        self.ctx.property_criteria = get_protocol(
            category='criteria', name=self.inputs.criteria.value
        )[self._PROPERTY_NAME]
        self.ctx.output_parameters['used_criteria'] = self.inputs.criteria.value

    def setup_code_resource_options(self):
        """
        setup resource options and parallelization for `PwCalculation` from inputs
        """
        if 'options' in self.inputs:
            self.ctx.options = self.inputs.options.get_dict()
        else:
            from aiida_sssp_workflow.utils import get_default_options

            self.ctx.options = get_default_options(with_mpi=True)

        if 'parallelization' in self.inputs:
            self.ctx.parallelization = self.inputs.parallelization.get_dict()
        else:
            self.ctx.parallelization = {}

    def run_reference(self):
        """
        run on reference calculation
        """
        ecutwfc = self.ctx.reference_ecutwfc
        ecutrho = ecutwfc * self.ctx.dual
        inputs = self._get_inputs(ecutwfc=round(ecutwfc), ecutrho=round(ecutrho))

        self.ctx.max_ecutrho = self.ctx.reference_ecutwfc * self.ctx.dual

        running = self.submit(self._EVALUATE_WORKCHAIN, **inputs)
        self.report(f'launching reference calculation: {running.process_label}<{running.pk}>')

        self.to_context(reference=running)

    def inspect_reference(self):
        try:
            workchain = self.ctx.reference
        except AttributeError as e:
            raise RuntimeError('Reference evaluation is not triggered') from e

        if not workchain.is_finished_ok:
            self.report(
                f'{workchain.process_label} pk={workchain.pk} for reference run is failed.'
            )
            return self.exit_codes.ERROR_REFERENCE_CALCULATION_FAILED

    def run_wfc_convergence_test(self):
        """
        run on all other evaluation sample points
        """
        for ecutwfc in self.ctx.ecutwfc_list[:-1]: # The last one is reference
            ecutrho = ecutwfc * self.ctx.dual
            ecutwfc, ecutrho = round(ecutwfc), round(ecutrho)
            inputs = self._get_inputs(ecutwfc=ecutwfc, ecutrho=ecutrho)

            running = self.submit(self._EVALUATE_WORKCHAIN, **inputs)
            self.report(
                f'launching fix ecutrho={ecutrho} [ecutwfc={ecutwfc}] {running.process_label}<{running.pk}>')

            self.to_context(children_wfc=append_(running))

    def inspect_wfc_convergence_test(self):
        # include reference node in the last
        sample_nodes = self.ctx.children_wfc + [self.ctx.reference]

        if 'extra_parameters' in self.ctx:
            output_parameters = self.result_general_process(
                self.ctx.reference,
                sample_nodes,
                extra_parameters=self.ctx.extra_parameters
            )
        else:
            output_parameters = self.result_general_process(
                self.ctx.reference, sample_nodes)

        self.out('output_parameters_wfc_test',
                 orm.Dict(dict=output_parameters).store())

        # specificly for precheck
        # always using precision criteria.
        # in this precheck, we run test with pre-fix dualkand scan the cutoff pairs
        # at ecutwfc equal to 150 Ry, 200 Ry and 300 Ry, with ecutwfc=300ry as reference.
        # Whether or not to run the subsequent convergence test depent on this precheck.
        # There are following four possibilities, two will abort and wait for further input:
        # 1. (Abort1: code=-300) Under 2 times strict criteria 200 ry not converged. Highest priority.
        # 2. (Abort2: code=-150) Under normal criteria 150 not converged w.r.t 300.
        # 3. (Good: code=200) All good as usual, 200 Ry used as reference.
        # 4. (Better: code=150) Even better, under 2 times strict criteria 150 Ry converged. Means
        # ecutwfc=150ry can be used as reference. But only give advice, in real run still use 200 Ry as
        # reference. This condition will accelerate calcualtion and will be used in aiidalab-sssp.
        if self.inputs.cutoff_control.value == 'precheck':
            precision_criteria = get_protocol(
                category='criteria', name='precision'
            )[self._PROPERTY_NAME]

            x = output_parameters['ecutwfc']
            # normal criteria
            y = [i for i in output_parameters[self._MEASURE_OUT_PROPERTY]]
            res_normal = convergence_analysis(orm.List(list=list(zip(x, y))),
                                    orm.Dict(dict=precision_criteria))
            # two time strict criteria
            y = [i * 2 for i in output_parameters[self._MEASURE_OUT_PROPERTY]]
            res_strict = convergence_analysis(orm.List(list=list(zip(x, y))),
                                    orm.Dict(dict=precision_criteria))

            if res_strict['cutoff'].value == 300:
                # 200 ry not converged
                self.ctx.output_parameters['precheck'] = {
                    'exit_status': -300,
                    'message': f"Damn, Super hard pseudo. Under 2 times strict criteria 200 ry not converged.",
                    'value': res_strict['value'].value,
                    'bounds': precision_criteria['bounds']
                }

            if res_strict['cutoff'].value == 200:
                # converged at 200 ry.
                self.ctx.output_parameters['precheck'] = {
                    'exit_status': 200,
                    'message': 'Good, 200 Ry should be used as reference.',
                    'value': res_strict['value'].value,
                    'bounds': precision_criteria['bounds']
                }

                if res_normal['cutoff'].value != 150:
                    # 150 not converged
                    self.ctx.output_parameters['precheck'] = {
                        'exit_status': -150,
                        'message': 'Bad, hard pseudo, 150 Ry not converged yet.',
                        'value': res_normal['value'].value,
                        'bounds': precision_criteria['bounds']
                    }

            if res_strict['cutoff'].value == 150:
                # converged at 150 ry.
                # However, this case is rathe useless, since 200 Ry already run and cached
                # using reference of 150 Ry has no improvement for the efficiency.
                self.ctx.output_parameters['precheck'] = {
                    'exit_status': 150,
                    'message': 'Better, 150 Ry can be used as reference.',
                    'value': res_strict['value'].value,
                    'bounds': precision_criteria['bounds']
                }


        criterias = get_protocol(category='criteria')
        all_criteria_wavefunction_cutoff = {}
        for name, criteria in criterias.items():
            property_criteria = criteria[self._PROPERTY_NAME]
            # from the fix dual result find the converge wfc cutoff
            x = output_parameters['ecutwfc']
            y = output_parameters[self._MEASURE_OUT_PROPERTY]
            res = convergence_analysis(orm.List(list=list(zip(x, y))),
                                    orm.Dict(dict=property_criteria))

            all_criteria_wavefunction_cutoff[name] = res['cutoff'].value

            # specificly write output for set criteria
            if name == self.inputs.criteria.value:
                self.ctx.output_parameters['wavefunction_cutoff'] = self.ctx.wfc_cutoff= res['cutoff'].value

                self.logger.info(
                    f"The wfc convergence at {self.ctx.wfc_cutoff} with value={res['value'].value}"
                )

        # write output wavefunction cutoff in all criteria.
        self.ctx.output_parameters['all_criteria_wavefunction_cutoff'] = all_criteria_wavefunction_cutoff

    def run_rho_convergence_test(self):
        """
        run rho converge test on fix wfc cutoff
        """
        ecutwfc = self.ctx.wfc_cutoff
        # Only run rho test when ecutrho less than the max reference
        # otherwise meaningless for the exceeding cutoff test
        for ecutrho in [dual * ecutwfc for dual in self.ctx.dual_scan_list]:
            ecutwfc, ecutrho = round(ecutwfc), round(ecutrho)
            # TODO check and assert that ecutrho should not exceed max_ecutrho
            inputs = self._get_inputs(ecutwfc=ecutwfc, ecutrho=ecutrho)

            running = self.submit(self._EVALUATE_WORKCHAIN, **inputs)
            self.report(
                f'launching fix ecutwfc={ecutwfc} [ecutrho={ecutrho}] {running.process_label}<{running.pk}>'
            )

            self.to_context(children_rho=append_(running))

    def inspect_rho_convergence_test(self):
        sample_nodes = self.ctx.children_rho + [self.ctx.reference]

        if 'extra_parameters' in self.ctx:
            output_parameters = self.result_general_process(
                self.ctx.reference,
                sample_nodes,
                extra_parameters=self.ctx.extra_parameters
            )
        else:
            output_parameters = self.result_general_process(
                self.ctx.reference, sample_nodes)

        self.out('output_parameters_rho_test',
                    orm.Dict(dict=output_parameters).store())

        # from the fix wfc cutoff result find the converge rho cutoff
        x = output_parameters['ecutrho']
        y = output_parameters[self._MEASURE_OUT_PROPERTY]
        res = convergence_analysis(orm.List(list=list(zip(x, y))),
                                    orm.Dict(dict=self.ctx.property_criteria))
        self.ctx.rho_cutoff, y_value = res['cutoff'].value, res[
            'value'].value
        self.ctx.output_parameters['chargedensity_cutoff'] = self.ctx.rho_cutoff

        self.logger.info(
            f'The rho convergence at {self.ctx.rho_cutoff} with value={y_value}'
        )

    def result_general_process(self, reference_node, sample_nodes, **kwargs) -> dict:
        """set results of sub-workflows to output ports"""
        children = sample_nodes
        success_children = [
            child for child in children if child.is_finished_ok
        ]

        ecutwfc_list = []
        ecutrho_list = []
        d_output_parameters = {}

        for key, value in self.get_result_metadata().items():
            d_output_parameters[key] = value

        for child_node in success_children:
            ecutwfc_list.append(child_node.outputs.ecutwfc.value)
            ecutrho_list.append(child_node.outputs.ecutrho.value)

            res = self.helper_compare_result_extract_fun(child_node,
                                                    reference_node, **kwargs)

            for key, value in res.items():
                if key not in self.get_result_metadata():
                    lst = d_output_parameters.get(key, [])
                    lst.append(value)
                    d_output_parameters.update({key: lst})

        d_output_parameters['ecutwfc'] = ecutwfc_list
        d_output_parameters['ecutrho'] = ecutrho_list

        return d_output_parameters

    def finalize(self):
        # store output_parameters
        self.out('output_parameters',
                 orm.Dict(dict=self.ctx.output_parameters).store())



# -*- coding: utf-8 -*-
"""
Convergence test on cohesive energy of a given pseudopotential
"""

from aiida import orm
from aiida.engine import calcfunction
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import RARE_EARTH_ELEMENTS, update_dict
from aiida_sssp_workflow.workflows.convergence._base import _BaseConvergenceWorkChain
from aiida_sssp_workflow.workflows.evaluate._delta import DeltaWorkChain

UpfData = DataFactory("pseudo.upf")


@calcfunction
def helper_delta_difference(
    input_parameters: orm.Dict, ref_parameters: orm.Dict
) -> orm.Dict:
    """calculate the delta difference from parameters"""
    res_delta = input_parameters["delta"]
    ref_delta = ref_parameters["delta"]
    relative_diff = abs((res_delta - ref_delta) / ref_delta) * 100

    res = {
        "delta": res_delta,
        "relative_diff": relative_diff,
        "delta_unit": "meV",
        "relative_unit": "%",
    }

    return orm.Dict(dict=res)


class ConvergenceDeltaWorkChain(_BaseConvergenceWorkChain):
    """WorkChain to converge test on delta factor of input structure"""

    # pylint: disable=too-many-instance-attributes

    _PROPERTY_NAME = "delta"
    _EVALUATE_WORKCHAIN = DeltaWorkChain
    _MEASURE_OUT_PROPERTY = "relative_diff"

    def init_setup(self):
        super().init_setup()
        self.ctx.extra_pw_parameters = {}

    def extra_setup_for_rare_earth_element(self):
        super().extra_setup_for_rare_earth_element()

    def setup_code_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        super().setup_code_parameters_from_protocol()

        # parse protocol
        protocol = self.ctx.protocol
        self._DEGAUSS = protocol["degauss"]
        self._OCCUPATIONS = protocol["occupations"]
        self._SMEARING = protocol["smearing"]
        self._CONV_THR = protocol["electron_conv_thr"]
        self._KDISTANCE = protocol["kpoints_distance"]

        self.ctx.scale_count = self._SCALE_COUNT = protocol["scale_count"]
        self.ctx.scale_increment = self._SCALE_INCREMENT = protocol["scale_increment"]

        # configuration for delta convergence
        if self.ctx.element in RARE_EARTH_ELEMENTS:
            self.ctx.configuration = "RE"
        else:
            self.ctx.configuration = "TYPICAL"

        # Set context parameters
        self.ctx.kpoints_distance = self._KDISTANCE
        self.ctx.pw_parameters = super()._get_pw_base_parameters(
            self._DEGAUSS, self._OCCUPATIONS, self._SMEARING, self._CONV_THR
        )

        self.logger.info(
            f"The pw parameters for convergence is: {self.ctx.pw_parameters}"
        )

    def _get_inputs(self, ecutwfc, ecutrho):
        """
        get inputs for the evaluation DeltaWorkChain by provide ecutwfc and ecutrho,
        all other parameters are fixed for the following steps
        """
        parameters = {
            "SYSTEM": {
                "ecutwfc": ecutwfc,
                "ecutrho": ecutrho,
            },
        }
        parameters = update_dict(parameters, self.ctx.pw_parameters)
        parameters["CONTROL"].pop(
            "tstress", None
        )  # this will rule this work chain out from caching

        # sparse kpoints and tetrahedra occupation in EOS reference calculation
        if self.ctx.element in RARE_EARTH_ELEMENTS:
            self.ctx.kpoints_distance = self._KDISTANCE + 0.05
            parameters["SYSTEM"].pop("smearing", None)
            parameters["SYSTEM"].pop("degauss", None)
            parameters["SYSTEM"]["occupations"] = "tetrahedra"

        inputs = {
            "eos": {
                "metadata": {"call_link_label": "delta_EOS"},
                "structure": self.ctx.structure,
                "kpoints_distance": orm.Float(self.ctx.kpoints_distance),
                "scale_count": orm.Int(self.ctx.scale_count),
                "scale_increment": orm.Float(self.ctx.scale_increment),
                "pw": {
                    "code": self.inputs.code,
                    "pseudos": self.ctx.pseudos,
                    "parameters": orm.Dict(dict=parameters),
                    "metadata": {
                        "options": self.ctx.options,
                    },
                    "parallelization": orm.Dict(dict=self.ctx.parallelization),
                },
            },
            "element": orm.Str(self.ctx.element),
            "configuration": orm.Str(self.ctx.configuration),
            "clean_workchain": self.inputs.clean_workchain,
        }

        return inputs

    def helper_compare_result_extract_fun(self, sample_node, reference_node, **kwargs):
        """extract"""
        sample_output = sample_node.outputs.output_parameters
        reference_output = reference_node.outputs.output_parameters

        res = helper_delta_difference(sample_output, reference_output).get_dict()

        return res

    def get_result_metadata(self):
        return {
            "delta_unit": "meV",
            "relative_unit": "%",
        }



# -*- coding: utf-8 -*-
"""utils.py"""

import collections.abc
import importlib
import json

import yaml
from aiida import orm
from aiida.engine import calcfunction
from aiida.plugins import DataFactory

UpfData = DataFactory("pseudo.upf")


def get_protocol(category, name=None):
    """Load and read protocol from faml file to a verbose dict
    if name not set, return whole protocol."""
    import_path = importlib.resources.path(
        "aiida_sssp_workflow.protocol", f"{category}.yml"
    )
    with import_path as pp_path, open(pp_path, "rb") as handle:
        protocol_dict = yaml.safe_load(
            handle
        )  # pylint: disable=attribute-defined-outside-init

    if name:
        return protocol_dict[name]
    else:
        return protocol_dict


RARE_EARTH_ELEMENTS = [
    "La",
    "Ce",
    "Pr",
    "Nd",
    "Pm",
    "Sm",
    "Eu",
    "Gd",
    "Tb",
    "Dy",
    "Ho",
    "Er",
    "Tm",
    "Yb",
    "Lu",
]

MAGNETIC_ELEMENTS = ["Mn", "O", "Cr", "Fe", "Co", "Ni"]

NONMETAL_ELEMENTS = [
    "H",
    "He",
    "B",
    "N",
    "O",
    "F",
    "Ne",
    "Si",
    "P",
    "Cl",
    "Ar",
    "Se",
    "Br",
    "Kr",
    "Te",
    "I",
    "Xe",
    "Rn",
]

HIGH_DUAL_ELEMENTS = ["O", "Fe", "Mn", "Hf", "Co", "Ni", "Cr"]

OXIDE_CONFIGURATIONS = ["XO", "XO2", "XO3", "X2O", "X2O3", "X2O5"]
UNARIE_CONFIGURATIONS = ["BCC", "FCC", "SC", "Diamond"]


def update_dict(d, u):
    """update dict by7 another dict, update include all hierarchy
    return the update dict instead of change the input dict"""
    # pylint: disable=invalid-name
    import copy

    ret = copy.deepcopy(d)
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            ret[k] = update_dict(ret.get(k, {}), v)
        else:
            ret[k] = v
    return ret


def get_standard_structure(
    element: str, prop: str, configuration=None
) -> orm.StructureData:
    """
    get cif abspath from element for bands measure and convergence
    property can be `delta`, `bands` or `convergence`

    The principles are for bands measure, using the configurations from Cottiner's paper since they are the groud state structures exist in real wolrd.
    And for lanthanides using the Nitrides from Wenzowech paper.
    Using the uniaries/diamond configurations for convergence verification.

    The structure is convert to primitive cell except for those magnetic elements
    configurations (typical ones of Cottiner's paper) for bands measure, since we need to
    set the starting magnetizations for sites.

    If prop is delta must provide specific configuration name.

    Args:
        element (str): element
        configuration (str): BCC, FCC, SC, Diamond, XO, XO2, XO3, X2O, X2O3, X2O5, RE

    Returns:
        orm.StructureData: return a orm.StructureData
    """
    # If for delta measure workflow
    base_cif_module = "aiida_sssp_workflow.statics.cif"
    if prop == "delta":
        if configuration == "RE":
            assert element in RARE_EARTH_ELEMENTS

            # use RE-nitrides of Wenzovich paper
            # from typical cif folder
            res_path = importlib.resources.path(
                f"{base_cif_module}.typical", f"{element}N.cif"
            )

        if configuration == "TYPICAL":
            res_path = importlib.resources.path(
                f"{base_cif_module}.typical", f"{element}.cif"
            )

        if configuration in OXIDE_CONFIGURATIONS:
            res_path = importlib.resources.path(
                f"{base_cif_module}.oxides", f"{element}_{configuration}.cif"
            )

        if configuration in UNARIE_CONFIGURATIONS:
            res_path = importlib.resources.path(
                f"{base_cif_module}.unaries", f"{element}_{configuration}.cif"
            )

    else:
        # If for bands measure or convergence workflow
        # after some back-forward, most elements only use typical structure
        # for bands and for convergence. But with the primitived structure.
        # But for future maintainance, I keep the mapping.json for configuration
        # mapping. Only At using unaries since it is not in typical.
        import_path = importlib.resources.path(
            "aiida_sssp_workflow.statics.cif", f"mapping.json"
        )

        with import_path as path, open(path, "r") as handle:
            mapping = json.load(handle)

        dir, fn = mapping[element][prop].split("/")
        res_path = importlib.resources.path(f"{base_cif_module}.{dir}", fn)

    if element in MAGNETIC_ELEMENTS:
        primitive_cell = False
    else:
        primitive_cell = True

    with res_path as path:
        structure = orm.CifData.get_or_create(str(path), use_first=True)[
            0
        ].get_structure(primitive_cell=primitive_cell)

    return structure


def parse_upf(upf_content: str) -> dict:
    """

    :param upf_content:
    :param check: if check the integrity of the pp file
    :return:
    """
    from upf_to_json import upf_to_json

    upf_dict = upf_to_json(upf_content, None)

    return upf_dict["pseudo_potential"]


def helper_parse_upf(upf: UpfData) -> dict:
    """parser upf"""
    header = parse_upf(upf.get_content())["header"]

    return header


def get_default_options(max_num_machines=1, max_wallclock_seconds=1800, with_mpi=False):
    """Return an instance of the options dictionary with the minimally required parameters for a `CalcJob`.

    :param max_num_machines: set the number of nodes, default=1
    :param max_wallclock_seconds: set the maximum number of wallclock seconds, default=1800
    :param with_mpi: whether to run the calculation with MPI enabled
    """
    return {
        "resources": {"num_machines": int(max_num_machines)},
        "max_wallclock_seconds": int(max_wallclock_seconds),
        "withmpi": with_mpi,
    }


def to_valid_key(name):
    """
    convert name into a valid key name which contain only alphanumeric and underscores
    """
    import re

    valid_name = re.sub(r"[^\w\s]", "_", name)

    return valid_name


@calcfunction
def convergence_analysis(xy: orm.List, criteria: orm.Dict):
    """
    xy is a list of xy tuple [(x1, y1), (x2, y2), ...] and
    criteria is a dict of {'mode': 'a', 'bounds': (0.0, 0.2)}
    """
    # sort xy
    sorted_xy = sorted(xy.get_list(), key=lambda k: k[0], reverse=True)
    criteria = criteria.get_dict()
    mode = criteria["mode"]

    cutoff, value = sorted_xy[0]
    if mode == 0:
        bounds = criteria["bounds"]
        eps = criteria["eps"]
        # from max cutoff, after some x all y is out of bound
        for x, y in sorted_xy:
            if bounds[0] - eps < y < bounds[1] + eps:
                cutoff, value = x, y
            else:
                break

    return {
        "cutoff": orm.Float(cutoff),
        "value": orm.Float(value),
    }


def reset_pseudos_for_magnetic(pseudo, structure):
    """
    override pseudos setting
    required for O, Mn, Cr typical configuration and
    diamond configuration of all magnetic elements where
    the kind names varies for sites
    """
    pseudos = {}
    for kind_name in structure.get_kind_names():
        pseudos[kind_name] = pseudo

    return pseudos


def get_magnetic_inputs(structure: orm.StructureData):
    """
    To set initial magnet to the magnetic system, need to set magnetic order to
    every magnetic element site, with certain pw starting_mainetization parameters.

    ! Only for typical configurations of magnetic elements.
    """
    MAG_INIT_Mn = {
        "Mn1": 0.5,
        "Mn2": -0.3,
        "Mn3": 0.5,
        "Mn4": -0.3,
    }  # pylint: disable=invalid-name
    MAG_INIT_O = {
        "O1": 0.5,
        "O2": 0.5,
        "O3": -0.5,
        "O4": -0.5,
    }  # pylint: disable=invalid-name
    MAG_INIT_Cr = {"Cr1": 0.5, "Cr2": -0.5}  # pylint: disable=invalid-name

    mag_structure = orm.StructureData(cell=structure.cell, pbc=structure.pbc)
    kind_name = structure.get_kind_names()[0]

    # ferromagnetic
    if kind_name in ["Fe", "Co", "Ni"]:
        for i, site in enumerate(structure.sites):
            mag_structure.append_atom(position=site.position, symbols=kind_name)

        parameters = {
            "SYSTEM": {
                "nspin": 2,
                "starting_magnetization": {kind_name: 0.2},
            },
        }

    #
    if kind_name in ["Mn", "O", "Cr"]:
        for i, site in enumerate(structure.sites):
            mag_structure.append_atom(
                position=site.position, symbols=kind_name, name=f"{kind_name}{i+1}"
            )

        if kind_name == "Mn":
            parameters = {
                "SYSTEM": {
                    "nspin": 2,
                    "starting_magnetization": MAG_INIT_Mn,
                },
            }

        if kind_name == "O":
            parameters = {
                "SYSTEM": {
                    "nspin": 2,
                    "starting_magnetization": MAG_INIT_O,
                },
            }

        if kind_name == "Cr":
            parameters = {
                "SYSTEM": {
                    "nspin": 2,
                    "starting_magnetization": MAG_INIT_Cr,
                },
            }

    return mag_structure, parameters



# -*- coding: utf-8 -*-
"""
Refactor from calcDelta.py v3.1 write by Kurt Lejaeghere
Copyright (C) 2012 Kurt Lejaeghere <Kurt.Lejaeghere@UGent.be>, Center for
Molecular Modeling (CMM), Ghent University, Ghent, Belgium
"""
import importlib
import json

import numpy as np
from aiida import orm
from aiida.engine import calcfunction

from aiida_sssp_workflow.calculations.wien2k_ref import WIEN2K_REF, WIEN2K_REN_REF
from aiida_sssp_workflow.utils import (
    OXIDE_CONFIGURATIONS,
    RARE_EARTH_ELEMENTS,
    UNARIE_CONFIGURATIONS,
)

# pylint: disable=invalid-name


def helper_get_v0_b0_b1(element: str, structure: str):
    """get eos reference of element"""
    import re

    if element in RARE_EARTH_ELEMENTS:
        element_str = f"{element}N"
    else:
        element_str = element

    regex = re.compile(
        rf"""{element_str}\s*
                        (?P<V0>\d*.\d*)\s*
                        (?P<B0>\d*.\d*)\s*
                        (?P<B1>\d*.\d*)""",
        re.VERBOSE,
    )
    if element not in RARE_EARTH_ELEMENTS:
        match = regex.search(WIEN2K_REF)
        V0 = match.group("V0")
        B0 = match.group("B0")
        B1 = match.group("B1")
    else:
        match = regex.search(WIEN2K_REN_REF)
        V0 = match.group("V0")
        B0 = match.group("B0")
        B1 = match.group("B1")

    echarge = 1.60217733e-19
    return float(V0), float(B0) / (echarge * 1.0e21), float(B1)


@calcfunction
def delta_analyze(element, configuration, V0, B0, B1, natoms) -> orm.Dict:
    """
    The calcfunction calculate the delta factor.
    return delta factor with unit (eV/atom)

    The configuration can be one of:
    - RE: for Rare-earth element
    - TYPICAL: for BM fit results from Corttiner's paper
    - One of OXIDES:
        - XO
        - XO2
        - XO3
        - X2O
        - X2O3
        - X2O5
    - One of UNARIES:
        - BCC
        - FCC
        - SC
        - Diamond

    conf_key is key in json file for configurations of every element.
    """
    element = element.value
    configuration = configuration.value
    V0 = V0.value
    B0 = B0.value
    B1 = B1.value
    natoms = natoms.value
    if configuration == "RE":
        assert element in RARE_EARTH_ELEMENTS

        ref_json = "WIEN2K_LANN.json"
        conf_key = f"{element}N"

    if configuration == "TYPICAL":
        ref_json = "WIEN2K_TYPICAL.json"
        conf_key = f"{element}"

    if configuration in UNARIE_CONFIGURATIONS:
        ref_json = "WIEN2K_UNARIES.json"
        conf_key = f"{element}-X/{configuration}"

    if configuration in OXIDE_CONFIGURATIONS:
        ref_json = "WIEN2K_OXIDES.json"
        conf_key = f"{element}-{configuration}"

    import_path = importlib.resources.path(
        "aiida_sssp_workflow.statics.AE_EOS", ref_json
    )
    with import_path as path, open(path, "rb") as handle:
        data = json.load(handle)

    BM_fit = data["BM_fit_data"][conf_key]
    ref_V0, ref_B0, ref_B1 = (
        BM_fit["min_volume"],
        BM_fit["bulk_modulus_ev_ang3"],
        BM_fit["bulk_deriv"],
    )

    # Delta computation
    delta, deltarel, delta1 = _calcDelta(ref_V0, ref_B0, ref_B1, V0, B0, B1)

    nu_measure = rel_errors_vec_length(
        ref_V0,
        ref_B0,
        ref_B1,
        V0,
        B0,
        B1,
        config_string=None,
        prefact=400,  # XXX: re-inspect this value and set it in protocol
        weight_b0=1 / 8,
        weight_b1=1 / 64,
    )

    return orm.Dict(
        dict={
            "delta": delta,
            "delta1": delta1,
            "delta_unit": "meV/atom",
            "delta_relative": deltarel,
            "delta_relative_unit": "%",
            "natoms": natoms,
            "delta/natoms": delta / natoms,
            "birch_murnaghan_results": [V0, B0, B1],
            "reference_wien2k_V0_B0_B1": [ref_V0, ref_B0, ref_B1],
            "V0_B0_B1_units_info": "eV/A^3 for B0",
            "rel_errors_vec_length": nu_measure,
            "nu/natoms": nu_measure / natoms,
        }
    )


def rel_errors_vec_length(
    v0w, b0w, b1w, v0f, b0f, b1f, config_string, prefact, weight_b0, weight_b1
):
    """
    Returns the length of the vector formed by the relative error of V0, B0, B1
    THE SIGNATURE OF THIS FUNCTION HAS BEEN CHOSEN TO MATCH THE ONE OF ALL THE OTHER FUNCTIONS
    RETURNING A QUANTITY THAT IS USEFUL FOR COMPARISON, THIS SIMPLIFIES THE CODE LATER.
    Even though config_string is not usd
    """
    V0err = 2 * (v0w - v0f) / (v0w + v0f)
    B0err = 2 * (b0w - b0f) / (b0w + b0f)
    B1err = 2 * (b1w - b1f) / (b1w + b1f)
    leng = np.sqrt(V0err**2 + (weight_b0 * B0err) ** 2 + (weight_b1 * B1err) ** 2)
    return leng * prefact


def _calcDelta(v0w, b0w, b1w, v0f, b0f, b1f, useasymm=False):
    """
    Calculate the Delta value, function copied from the official DeltaTest repository.
    I don't understand what it does, but it works.
    THE SIGNATURE OF THIS FUNCTION HAS BEEN CHOSEN TO MATCH THE ONE OF ALL THE OTHER FUNCTIONS
    RETURNING A QUANTITY THAT IS USEFUL FOR COMPARISON, THIS SIMPLIFIES THE CODE LATER.
    Even though 'config_string' is useless here.
    """
    # pylint: disable=too-many-statements, consider-using-enumerate

    vref = 30.0
    bref = 100.0 * 10.0**9.0 / 1.602176565e-19 / 10.0**30.0

    if useasymm:
        Vi = 0.94 * v0w
        Vf = 1.06 * v0w
    else:
        Vi = 0.94 * (v0w + v0f) / 2.0
        Vf = 1.06 * (v0w + v0f) / 2.0

    a3f = 9.0 * v0f**3.0 * b0f / 16.0 * (b1f - 4.0)
    a2f = 9.0 * v0f ** (7.0 / 3.0) * b0f / 16.0 * (14.0 - 3.0 * b1f)
    a1f = 9.0 * v0f ** (5.0 / 3.0) * b0f / 16.0 * (3.0 * b1f - 16.0)
    a0f = 9.0 * v0f * b0f / 16.0 * (6.0 - b1f)

    a3w = 9.0 * v0w**3.0 * b0w / 16.0 * (b1w - 4.0)
    a2w = 9.0 * v0w ** (7.0 / 3.0) * b0w / 16.0 * (14.0 - 3.0 * b1w)
    a1w = 9.0 * v0w ** (5.0 / 3.0) * b0w / 16.0 * (3.0 * b1w - 16.0)
    a0w = 9.0 * v0w * b0w / 16.0 * (6.0 - b1w)

    x = [0, 0, 0, 0, 0, 0, 0]

    x[0] = (a0f - a0w) ** 2
    x[1] = 6.0 * (a1f - a1w) * (a0f - a0w)
    x[2] = -3.0 * (2.0 * (a2f - a2w) * (a0f - a0w) + (a1f - a1w) ** 2.0)
    x[3] = -2.0 * (a3f - a3w) * (a0f - a0w) - 2.0 * (a2f - a2w) * (a1f - a1w)
    x[4] = -3.0 / 5.0 * (2.0 * (a3f - a3w) * (a1f - a1w) + (a2f - a2w) ** 2.0)
    x[5] = -6.0 / 7.0 * (a3f - a3w) * (a2f - a2w)
    x[6] = -1.0 / 3.0 * (a3f - a3w) ** 2.0

    y = [0, 0, 0, 0, 0, 0, 0]

    y[0] = (a0f + a0w) ** 2 / 4.0
    y[1] = 3.0 * (a1f + a1w) * (a0f + a0w) / 2.0
    y[2] = -3.0 * (2.0 * (a2f + a2w) * (a0f + a0w) + (a1f + a1w) ** 2.0) / 4.0
    y[3] = -(a3f + a3w) * (a0f + a0w) / 2.0 - (a2f + a2w) * (a1f + a1w) / 2.0
    y[4] = -3.0 / 20.0 * (2.0 * (a3f + a3w) * (a1f + a1w) + (a2f + a2w) ** 2.0)
    y[5] = -3.0 / 14.0 * (a3f + a3w) * (a2f + a2w)
    y[6] = -1.0 / 12.0 * (a3f + a3w) ** 2.0

    Fi = np.zeros_like(Vi)
    Ff = np.zeros_like(Vf)

    Gi = np.zeros_like(Vi)
    Gf = np.zeros_like(Vf)

    for n in range(7):
        Fi = Fi + x[n] * Vi ** (-(2.0 * n - 3.0) / 3.0)
        Ff = Ff + x[n] * Vf ** (-(2.0 * n - 3.0) / 3.0)

        Gi = Gi + y[n] * Vi ** (-(2.0 * n - 3.0) / 3.0)
        Gf = Gf + y[n] * Vf ** (-(2.0 * n - 3.0) / 3.0)

    Delta = 1000.0 * np.sqrt((Ff - Fi) / (Vf - Vi))
    Deltarel = 100.0 * np.sqrt((Ff - Fi) / (Gf - Gi))
    if useasymm:
        Delta1 = 1000.0 * np.sqrt((Ff - Fi) / (Vf - Vi)) / v0w / b0w * vref * bref
    else:
        Delta1 = (
            1000.0
            * np.sqrt((Ff - Fi) / (Vf - Vi))
            / (v0w + v0f)
            / (b0w + b0f)
            * 4.0
            * vref
            * bref
        )

    return Delta, Deltarel, Delta1



# -*- coding: utf-8 -*-
"""dump verification result"""
import json

import click
from aiida.common import AttributeDict
from aiida.orm import Dict, load_node

from aiida_sssp_workflow.cli import cmd_root


def flatten_output(attr_dict, node_collection):
    """
    flaten output dict node
    node_collection is a list to accumulate the nodes that not unfolded

    For some namespace do not unfold dict
    - band_parameters

    For output nodes not being expanded, write down the uuid for making archive
    """
    do_not_unfold = ["band_parameters", "scf_parameters", "seekpath_parameters"]

    for key, value in attr_dict.items():
        if isinstance(value, AttributeDict):
            # keep on unfold
            flatten_output(value, node_collection)
        elif isinstance(value, Dict) and key not in do_not_unfold:
            attr_dict[key] = value.get_dict()
        else:
            # node type not handled attach uuid
            attr_dict[key] = value.uuid
            node_collection.append(value.uuid)

    # print(archive_uuids)
    return attr_dict


@cmd_root.command("dump")
@click.argument("nodes", type=str, nargs=-1)
@click.argument("filename", type=click.Path())
@click.argument("archive", type=click.Path())
def dump_output(nodes, filename, archive):
    """dump the verification result"""
    from aiida.tools.importexport import ExportFileFormat, export

    res = {}
    archive_uuids = []
    for node in nodes:
        _node = load_node(node)
        label = _node.extras.get("label")
        attr_dict = _node.outputs._construct_attribute_dict(incoming=False)

        res[label] = flatten_output(attr_dict, archive_uuids)

    with open(filename, "w") as fh:
        json.dump(dict(res), fh, indent=2)

    kwargs = {
        "input_calc_forward": False,
        "input_work_forward": False,
        "create_backward": False,
        "return_backward": False,
        "call_calc_backward": False,
        "call_work_backward": False,
        "include_comments": False,
        "include_logs": False,
        "overwrite": True,
    }

    # print(archive_uuids)
    export(
        [load_node(i) for i in archive_uuids],
        filename=archive,
        file_format=ExportFileFormat.ZIP,
        **kwargs
    )



"""Base Measure Workchain since bands measure and delta measure share same input ports
This class makes it able to expose delta measure and bands measure inputs to verification
workchain.
"""
from aiida import orm
from aiida.plugins import DataFactory

from aiida_sssp_workflow.workflows import SelfCleanWorkChain

UpfData = DataFactory("pseudo.upf")


class _BaseMeasureWorkChain(SelfCleanWorkChain):
    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.input('code', valid_type=orm.Code,
                    help='The `pw.x` code use for the `PwCalculation`.')
        spec.input('pseudo', valid_type=UpfData, required=True,
                    help='Pseudopotential to be verified')
        spec.input('protocol', valid_type=orm.Str, required=True,
                    help='The protocol which define input calculation parameters.')
        spec.input('cutoff_control', valid_type=orm.Str, default=lambda: orm.Str('standard'),
                    help='The control protocol where define max_wfc.')
        spec.input('options', valid_type=orm.Dict, required=True,
                    help='Optional `options` to use for the `PwCalculations`.')
        spec.input('parallelization', valid_type=orm.Dict, required=True,
                    help='Parallelization options for the `PwCalculations`.')



# -*- coding: utf-8 -*-
"""
Bands distance of many input pseudos
"""

from aiida import orm
from aiida.engine import ToContext, if_
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import (
    HIGH_DUAL_ELEMENTS,
    MAGNETIC_ELEMENTS,
    NONMETAL_ELEMENTS,
    RARE_EARTH_ELEMENTS,
    get_magnetic_inputs,
    get_protocol,
    get_standard_structure,
    reset_pseudos_for_magnetic,
    update_dict,
)
from aiida_sssp_workflow.workflows.common import (
    get_extra_parameters_for_lanthanides,
    get_pseudo_element_and_type,
    get_pseudo_N,
)
from aiida_sssp_workflow.workflows.evaluate._bands import BandsWorkChain
from aiida_sssp_workflow.workflows.measure import _BaseMeasureWorkChain

UpfData = DataFactory('pseudo.upf')


def validate_input_pseudos(d_pseudos, _):
    """Validate that all input pseudos map to same element"""
    element = set(pseudo.element for pseudo in d_pseudos.values())

    if len(element) > 1:
        return f'The pseudos corespond to different elements {element}.'


class BandsMeasureWorkChain(_BaseMeasureWorkChain):
    """
    WorkChain to run bands measure,
    run without sym for distance compare and band structure along the path
    """
    _RY_TO_EV = 13.6056980659

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)

        spec.outline(
            cls.init_setup,
            if_(cls.is_magnetic_element)(
                cls.extra_setup_for_magnetic_element,
            ),
            if_(cls.is_rare_earth_element)(
                cls.extra_setup_for_rare_earth_element, ),
            cls.setup_pw_parameters_from_protocol,
            cls.run_bands_evaluation,
            cls.finalize,
        )

        spec.expose_outputs(BandsWorkChain)

    def init_setup(self):
        """
        This step contains all preparation before actaul setup, e.g. set
        the context of element, base_structure, base pw_parameters and pseudos.
        """
        # parse pseudo and output its header information
        self.ctx.pw_parameters = {}

        element = self.ctx.element = self.inputs.pseudo.element
        self.ctx.pseudos = {self.ctx.element: self.inputs.pseudo}

        self.ctx.structure = get_standard_structure(
            element,
            prop="bands",
        )

        # extra setting for bands convergence
        self.ctx.is_metal = element not in NONMETAL_ELEMENTS

    def is_magnetic_element(self):
        """Check if the element is magnetic"""
        return self.ctx.element in MAGNETIC_ELEMENTS

    def extra_setup_for_magnetic_element(self):
        """Extra setup for magnetic element"""
        self.ctx.structure, magnetic_extra_parameters = get_magnetic_inputs(self.ctx.structure)
        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters, magnetic_extra_parameters)

        # override pseudos setting
        # required for O, Mn, Cr where the kind names varies for sites
        self.ctx.pseudos = reset_pseudos_for_magnetic(self.inputs.pseudo, self.ctx.structure)


    def is_rare_earth_element(self):
        """Check if the element is rare earth"""
        return self.ctx.element in RARE_EARTH_ELEMENTS

    def extra_setup_for_rare_earth_element(self):
        """Extra setup for rare earth element"""
        nbnd_factor = 2.0
        pseudo_N = get_pseudo_N()
        self.ctx.pseudos['N'] = pseudo_N
        pseudo_RE = self.inputs.pseudo
        nbnd = nbnd_factor * (pseudo_N.z_valence + pseudo_RE.z_valence)
        pw_parameters = get_extra_parameters_for_lanthanides(self.ctx.element, nbnd)

        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters, pw_parameters)

    def setup_pw_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        protocol = get_protocol(category="bands", name=self.inputs.protocol.value)
        self._DEGAUSS = protocol['degauss']
        self._OCCUPATIONS = protocol['occupations']
        self._SMEARING = protocol['smearing']
        self._CONV_THR = protocol['electron_conv_thr']

        self._INIT_NBANDS_FACTOR = protocol['init_nbands_factor']
        self._FERMI_SHIFT = protocol['fermi_shift']

        cutoff_control = get_protocol(
            category="control", name=self.inputs.cutoff_control.value
        )
        self._ECUTWFC = cutoff_control["max_wfc"]

        self.ctx.ecutwfc = self._ECUTWFC
        self.ctx.init_nbands_factor = self._INIT_NBANDS_FACTOR
        self.ctx.fermi_shift = self._FERMI_SHIFT

        self.ctx.kpoints_distance_scf = protocol['kpoints_distance_scf']
        self.ctx.kpoints_distance_bands = protocol['kpoints_distance_bands']
        self.ctx.kpoints_distance_band_structure = protocol['kpoints_distance_band_structure']


        parameters = {
            "SYSTEM": {
                "degauss": self._DEGAUSS,
                "occupations": self._OCCUPATIONS,
                "smearing": self._SMEARING,
            },
            "ELECTRONS": {
                "conv_thr": self._CONV_THR,
            },
        }

        self.ctx.ecutwfc = self._ECUTWFC

        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters, parameters)

        self.logger.info(
            f'The pw parameters for convergence is: {self.ctx.pw_parameters}'
        )

    def _get_inputs(self, pseudos):
        """
        get inputs for the bands evaluation with given pseudo
        """
        element, pseudo_type = get_pseudo_element_and_type(self.inputs.pseudo)
        if pseudo_type in ['NC', 'SL']:
            ecutrho = self.ctx.ecutwfc * 4
        else:
            ecutrho = self.ctx.ecutwfc * 8

        if element in HIGH_DUAL_ELEMENTS and pseudo_type not in ['NC', 'SL']:
            ecutrho = self.ctx.ecutwfc * 18

        if element in RARE_EARTH_ELEMENTS:
            # since nitrides is used, the pseudo of N is non-NC
            # The N.us.z_5.ld1.theose.v0 is used so set dual equal to 8
            ecutrho = self.ctx.ecutwfc * 8

        parameters = {
            "SYSTEM": {
                "ecutwfc": round(self.ctx.ecutwfc),
                "ecutrho": round(ecutrho),
            },
        }
        parameters = update_dict(parameters, self.ctx.pw_parameters)

        parameters_bands = update_dict(parameters, {})
        parameters_bands["SYSTEM"].pop("nbnd", None)

        inputs = {
            "structure": self.ctx.structure,
            "scf": {
                "pw": {
                    "code": self.inputs.code,
                    "pseudos": pseudos,
                    "parameters": orm.Dict(dict=parameters),
                    "metadata": {
                        "options": self.inputs.options.get_dict(),
                    },
                    "parallelization": self.inputs.parallelization,
                },
                "kpoints_distance": orm.Float(self.ctx.kpoints_distance_scf),
            },
            "bands": {
                "pw": {
                    "code": self.inputs.code,
                    "pseudos": pseudos,
                    "parameters": orm.Dict(dict=parameters_bands),
                    "metadata": {
                        "options": self.inputs.options.get_dict(),
                    },
                    "parallelization": self.inputs.parallelization,
                },
            },
            "kpoints_distance_bands": orm.Float(self.ctx.kpoints_distance_bands),
            "init_nbands_factor": orm.Float(self.ctx.init_nbands_factor),
            "fermi_shift": orm.Float(self.ctx.fermi_shift),
            "run_bands_structure": orm.Bool(True),
            "kpoints_distance_band_structure": orm.Float(self.ctx.kpoints_distance_band_structure),
            "clean_workchain": self.inputs.clean_workchain,
        }

        return inputs

    def run_bands_evaluation(self):
        """run bands evaluation of psp in inputs list"""
        inputs = self._get_inputs(self.ctx.pseudos)

        running = self.submit(BandsWorkChain, **inputs)

        self.report(
            f'launching BandsWorkChain<{running.pk}>'
        )

        return ToContext(bands=running)

    def finalize(self):
        """inspect bands run results"""
        self.out_many(
            self.exposed_outputs(self.ctx.bands, BandsWorkChain)
        )



# -*- coding: utf-8 -*-
"""
Convergence test on cohesive energy of a given pseudopotential
"""

from aiida import orm
from aiida.engine import calcfunction
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import RARE_EARTH_ELEMENTS, update_dict
from aiida_sssp_workflow.workflows.convergence._base import _BaseConvergenceWorkChain
from aiida_sssp_workflow.workflows.evaluate._cohesive_energy import (
    CohesiveEnergyWorkChain,
)

UpfData = DataFactory("pseudo.upf")


@calcfunction
def helper_cohesive_energy_difference(
    input_parameters: orm.Dict, ref_parameters: orm.Dict
) -> orm.Dict:
    """calculate the cohesive energy difference from parameters"""
    res_energy = input_parameters["cohesive_energy_per_atom"]
    ref_energy = ref_parameters["cohesive_energy_per_atom"]
    absolute_diff = abs(res_energy - ref_energy) * 1000.0
    relative_diff = abs((res_energy - ref_energy) / ref_energy) * 100

    res = {
        "cohesive_energy_per_atom": res_energy,
        "absolute_diff": absolute_diff,
        "relative_diff": relative_diff,
        "absolute_unit": "meV/atom",
        "relative_unit": "%",
    }

    return orm.Dict(dict=res)


class ConvergenceCohesiveEnergyWorkChain(_BaseConvergenceWorkChain):
    """WorkChain to converge test on cohisive energy of input structure"""

    # pylint: disable=too-many-instance-attributes

    _PROPERTY_NAME = "cohesive_energy"
    _EVALUATE_WORKCHAIN = CohesiveEnergyWorkChain
    _MEASURE_OUT_PROPERTY = "absolute_diff"

    def init_setup(self):
        super().init_setup()
        self.ctx.extra_pw_parameters = {}
        self.ctx.extra_pw_parameters_for_atom = {}

    def extra_setup_for_magnetic_element(self):
        """Extra setup for magnetic element, for atom especially"""
        super().extra_setup_for_magnetic_element()
        self.ctx.extra_pw_parameters_for_atom = {
            self.ctx.element: {
                "SYSTEM": {
                    "nspin": 2,
                    "starting_magnetization": {
                        self.ctx.element: 0.5,
                    },
                },
                "ELECTRONS": {
                    "diagonalization": "cg",
                    "mixing_beta": 0.5,
                    "electron_maxstep": 200,
                },
            }
        }

    def extra_setup_for_rare_earth_element(self):
        """Extra setup for rare earth element, for atom especially"""
        super().extra_setup_for_rare_earth_element()
        self.ctx.extra_pw_parameters_for_atom = {
            self.ctx.element: {
                "SYSTEM": {
                    "nspin": 2,
                    "starting_magnetization": {
                        self.ctx.element: 0.5,
                    },
                    # Need high number of bands to make atom calculation of lanthanoids
                    # converged.
                    "nbnd": int(self.inputs.pseudo.z_valence * 3),
                },
                "ELECTRONS": {
                    "diagonalization": "cg",
                    "mixing_beta": 0.3,  # even small mixing_beta value
                    "electron_maxstep": 200,
                },
            },
        }

    def setup_code_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        super().setup_code_parameters_from_protocol()

        # parse protocol
        protocol = self.ctx.protocol
        self._DEGAUSS = protocol["degauss"]
        self._OCCUPATIONS = protocol["occupations"]
        self._BULK_SMEARING = protocol["smearing"]
        self._ATOM_SMEARING = protocol["atom_smearing"]
        self._CONV_THR = protocol["electron_conv_thr"]
        self.ctx.kpoints_distance = self._KDISTANCE = protocol["kpoints_distance"]
        self.ctx.vacuum_length = self._VACUUM_LENGTH = protocol["vacuum_length"]

        # Set context parameters
        self.ctx.bulk_parameters = super()._get_pw_base_parameters(
            self._DEGAUSS, self._OCCUPATIONS, self._BULK_SMEARING, self._CONV_THR
        )
        base_atom_pw_parameters = {
            "SYSTEM": {
                "degauss": self._DEGAUSS,
                "occupations": self._OCCUPATIONS,
                "smearing": self._ATOM_SMEARING,
            },
            "ELECTRONS": {
                "conv_thr": self._CONV_THR,
            },
        }

        self.ctx.atom_parameters = {}
        for element in self.ctx.structure.get_symbols_set():
            self.ctx.atom_parameters[element] = base_atom_pw_parameters

        self.ctx.atom_parameters = update_dict(
            self.ctx.atom_parameters, self.ctx.extra_pw_parameters_for_atom
        )

    def _get_inputs(self, ecutwfc, ecutrho):
        """
        get inputs for the evaluation CohesiveWorkChain by provide ecutwfc and ecutrho,
        all other parameters are fixed for the following steps
        """
        update_parameters = {
            "SYSTEM": {
                "ecutwfc": ecutwfc,
                "ecutrho": ecutrho,
            }
        }
        bulk_parameters = update_dict(self.ctx.bulk_parameters, update_parameters)

        atom_kpoints = orm.KpointsData()
        atom_kpoints.set_kpoints_mesh([1, 1, 1])

        # atomic parallelization always set npool to 1 since only one kpoints
        # requires no k parallel
        atomic_parallelization = update_dict(self.ctx.parallelization, {})
        atomic_parallelization.pop("npool", None)
        atomic_parallelization.pop("ndiag", None)
        atomic_parallelization = update_dict(atomic_parallelization, {"npool": 1})
        atomic_parallelization = update_dict(atomic_parallelization, {"ndiag": 1})

        # atomic option if mpiprocs too many confine it too no larger than 32 procs
        atomic_options = update_dict(self.ctx.options, {})
        if atomic_options["resources"]["num_mpiprocs_per_machine"] > 32:
            # copy is a shallow copy, so using update_dict.
            # if simply assign the value will change also the original dict
            atomic_options = update_dict(atomic_options, {"resources": {"num_mpiprocs_per_machine": 32}})

        # atomic calculation for lanthanides require more time to finish.
        if self.ctx.element in RARE_EARTH_ELEMENTS:
            pw_max_walltime = self.ctx.options.get("max_wallclock_seconds", None)
            if pw_max_walltime:
                atomic_options["max_wallclock_seconds"] = pw_max_walltime * 4

        # atom_parameters update with ecutwfc and ecutrho
        atom_parameters = update_dict(self.ctx.atom_parameters, {})
        for element in atom_parameters.keys():
            atom_parameters[element] = update_dict(atom_parameters[element], update_parameters)

        inputs = {
            "pseudos": self.ctx.pseudos,
            "structure": self.ctx.structure,
            "atom_parameters": orm.Dict(dict=atom_parameters),
            "vacuum_length": orm.Float(self.ctx.vacuum_length),
            "bulk": {
                "metadata": {"call_link_label": "bulk_scf"},
                "pw": {
                    "code": self.inputs.code,
                    "parameters": orm.Dict(dict=bulk_parameters),
                    "metadata": {
                        "options": self.ctx.options,
                    },
                    "parallelization": orm.Dict(dict=self.ctx.parallelization),
                },
                "kpoints_distance": orm.Float(self.ctx.kpoints_distance),
            },
            "atom": {
                "metadata": {"call_link_label": "atom_scf"},
                "pw": {
                    "code": self.inputs.code,
                    "parameters": orm.Dict(dict={}),
                    "metadata": {
                        "options": atomic_options,
                    },
                    "parallelization": orm.Dict(dict=atomic_parallelization),
                },
                "kpoints": atom_kpoints,
            },
            "clean_workchain": self.inputs.clean_workchain,
        }

        return inputs

    def helper_compare_result_extract_fun(self, sample_node, reference_node, **kwargs):
        """extract"""
        sample_output = sample_node.outputs.output_parameters
        reference_output = reference_node.outputs.output_parameters

        res = helper_cohesive_energy_difference(
            sample_output, reference_output
        ).get_dict()

        return res

    def get_result_metadata(self):
        return {
            "absolute_unit": "meV/atom",
            "relative_unit": "%",
        }



# -*- coding: utf-8 -*-
"""
A calcfunctian create_isolate_atom
Create the structure of isolate atom
"""
from aiida import orm
from aiida.engine import ToContext, calcfunction
from aiida.plugins import DataFactory, WorkflowFactory

from . import _BaseEvaluateWorkChain

PwBaseWorkflow = WorkflowFactory("quantumespresso.pw.base")
UpfData = DataFactory("pseudo.upf")


@calcfunction
def helper_get_hydrostatic_stress(output_trajectory, output_parameters):
    """
    doc
    """
    import numpy as np

    output_stress = output_trajectory.get_array("stress")[0]
    stress_unit = output_parameters["stress_units"]
    hydrostatic_stress = np.trace(output_stress) / 3.0
    return orm.Dict(
        dict={
            "stress_unit": stress_unit,
            "hydrostatic_stress": hydrostatic_stress,
        }
    )


class PressureWorkChain(_BaseEvaluateWorkChain):
    """WorkChain to calculate cohisive energy of input structure"""

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.expose_inputs(PwBaseWorkflow, include=['pw', 'kpoints_distance'])

        spec.outline(
            cls.run_scf,
            cls.inspect_scf,
            cls.finalize,
        )
        spec.output('output_parameters', valid_type=orm.Dict, required=True,
                    help='The output parameters include pressure of the structure.')

        # yapf: enable

    def run_scf(self):
        """
        set the inputs and submit scf
        """
        inputs = self.exposed_inputs(PwBaseWorkflow)

        running = self.submit(PwBaseWorkflow, **inputs)
        self.report(f"Running pw calculation pk={running.pk}")
        return ToContext(workchain_scf=running)

    def inspect_scf(self):
        """inspect the result of scf calculation."""
        workchain = self.ctx.workchain_scf

        if not workchain.is_finished_ok:
            self.report(
                f"PwBaseWorkChain for pressure evaluation failed with exit status {workchain.exit_status}"
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED_SCF

        output_trajectory = workchain.outputs.output_trajectory
        output_parameters = workchain.outputs.output_parameters

        # Return the output parameters of current workchain
        output_parameters = helper_get_hydrostatic_stress(
            output_trajectory, output_parameters
        )
        self.ctx.ecutwfc = workchain.inputs.pw.parameters["SYSTEM"]["ecutwfc"]
        self.ctx.ecutrho = workchain.inputs.pw.parameters["SYSTEM"]["ecutrho"]
        self.out("output_parameters", output_parameters)

    def finalize(self):
        """set ecutwfc and ecutrho"""
        self.out("ecutwfc", orm.Int(self.ctx.ecutwfc).store())
        self.out("ecutrho", orm.Int(self.ctx.ecutrho).store())



# -*- coding: utf-8 -*-
"""
WorkChain calculate phonon frequencies at Gamma
"""
from aiida import orm
from aiida.common import NotExistentAttributeError
from aiida.engine import ToContext, while_
from aiida.plugins import DataFactory, WorkflowFactory

from . import _BaseEvaluateWorkChain

PwBaseWorkflow = WorkflowFactory("quantumespresso.pw.base")
PhBaseWorkflow = WorkflowFactory("quantumespresso.ph.base")
UpfData = DataFactory("pseudo.upf")


class PhononFrequenciesWorkChain(_BaseEvaluateWorkChain):
    """WorkChain to calculate cohisive energy of input structure"""

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.expose_inputs(PwBaseWorkflow, namespace='scf', include=['metadata', 'pw', 'kpoints_distance'])
        spec.expose_inputs(PhBaseWorkflow, namespace='phonon', exclude=['ph.parent_folder'])

        spec.outline(
            cls.setup,
            while_(cls.is_pw_not_ready_for_ph)(
                cls.run_scf,
                cls.inspect_scf,
            ),
            cls.run_ph,
            cls.inspect_ph,
            cls.finalize,
        )
        spec.output('output_parameters', valid_type=orm.Dict, required=True,
                    help='The output parameters include phonon frequencies.')
        spec.exit_code(211, 'ERROR_NO_REMOTE_FOLDER',
                    message='The remote folder node not exist')
        spec.exit_code(202, 'ERROR_SUB_PROCESS_FAILED_PH',
                    message='The `PhBaseWorkChain` sub process failed.')
        # yapf: enable

    def setup(self):
        self.ctx.not_ready_for_ph = True

    def is_pw_not_ready_for_ph(self):
        """used to check if the remote folder is not empty, otherwise rerun pw with caching off"""
        return self.ctx.not_ready_for_ph

    def run_scf(self):
        """
        set the inputs and submit scf
        """
        inputs = self.exposed_inputs(PwBaseWorkflow, namespace="scf")

        running = self.submit(PwBaseWorkflow, **inputs)
        self.report(f"Running pw calculation pk={running.pk}")
        return ToContext(workchain_scf=running)

    def inspect_scf(self):
        """inspect the result of scf calculation if fail do not continue."""
        workchain = self.ctx.workchain_scf

        if not workchain.is_finished_ok:
            self.logger.warning(
                f"PwBaseWorkChain failed with exit status {workchain.exit_status}"
            )
            # set condition to False to break loop
            self.ctx.not_ready_for_ph = False

            return self.exit_codes.ERROR_SUB_PROCESS_FAILED_SCF

        try:
            remote_folder = self.ctx.scf_remote_folder = workchain.outputs.remote_folder

            if not remote_folder.is_empty:
                # when the remote_folder is not empty we regard it is ready for ph
                # even if the subsequent ph is successful
                self.ctx.not_ready_for_ph = False
            else:
                # set all same node to caching off and rerun
                pw_node = [
                    c for c in workchain.called if isinstance(c, orm.CalcJobNode)
                ][0]
                all_same_nodes = pw_node.get_all_same_nodes()
                for node in all_same_nodes:
                    node.delete_extra("_aiida_hash")

        except NotExistentAttributeError:
            # set condition to False to break loop
            self.ctx.not_ready_for_ph = False
            return self.exit_codes.ERROR_NO_REMOTE_FOLDER

        self.ctx.ecutwfc = workchain.inputs.pw.parameters["SYSTEM"]["ecutwfc"]
        self.ctx.ecutrho = workchain.inputs.pw.parameters["SYSTEM"]["ecutrho"]

        self.ctx.calc_time = workchain.outputs.output_parameters["wall_time_seconds"]

    def run_ph(self):
        """
        set the inputs and submit ph calculation to get quantities for phonon evaluation
        """

        inputs = self.exposed_inputs(PhBaseWorkflow, namespace="phonon")
        inputs["ph"]["parent_folder"] = self.ctx.scf_remote_folder

        running = self.submit(PhBaseWorkflow, **inputs)
        self.report(f"Running ph calculation pk={running.pk}")
        return ToContext(workchain_ph=running)

    def inspect_ph(self):
        """inspect the result of ph calculation."""
        workchain = self.ctx.workchain_ph

        if not workchain.is_finished_ok:
            self.report(
                f"PhBaseWorkChain for pressure evaluation failed with exit status {workchain.exit_status}"
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED_PH

        self.ctx.calc_time += workchain.outputs.output_parameters["wall_time_seconds"]
        output_parameters = workchain.outputs.output_parameters.get_dict()
        output_parameters.update(
            {
                "total_calc_time": self.ctx.calc_time,
                "time_unit": "s",
            }
        )
        self.out("output_parameters", orm.Dict(dict=output_parameters).store())

    def finalize(self):
        """set ecutwfc and ecutrho"""
        self.out("ecutwfc", orm.Int(self.ctx.ecutwfc).store())
        self.out("ecutrho", orm.Int(self.ctx.ecutrho).store())



# -*- coding: utf-8 -*-
"""
calculate bands distance
"""
import numpy as np
from aiida import orm

from aiida_sssp_workflow.efermi import find_efermi


def get_homo(bands, num_electrons: int):
    """
    This function only work for insulator,
    therefore the num_electrons is even number.
    """
    assert num_electrons % 2 == 0, f"There are {num_electrons} electrons, must be metal"
    # get homo band
    band = bands[:, num_electrons // 2 - 1]
    return max(band)


def fermi_dirac(band_energy, fermi_energy, smearing):
    """
    The first argument can be an array
    """
    old_settings = np.seterr(over="raise", divide="raise")
    try:
        res = 1.0 / (np.exp((band_energy - fermi_energy) / smearing) + 1.0)
    except FloatingPointError:
        res = np.heaviside(fermi_energy - band_energy, 1.0)
    np.seterr(**old_settings)

    return res


def retrieve_bands(
    bandsdata: orm.BandsData,
    start_band,
    num_electrons,
    efermi,
    smearing,
    do_smearing,
):
    """
    collect the bands of certain number with setting the start band.
    In order to make sure that when comparing two bands distance the number of bands is the same.

    The bands calculation of magnetic elements will giving a three dimensional bands where the
    first dimension is for the up and down spin.
    I simply concatenate along the first dimension.
    """
    bands = bandsdata.get_bands()
    kpoints, weights = bandsdata.get_kpoints(also_weights=True)

    # reduce by first dimension of up, down spins
    if len(bands.shape) > 2:
        nspin, nk, nbands = bands.shape
        bands = bands.reshape(nk, nbands * nspin)
        np.sort(bands)  # sort along last axis - eigenvalues of specific kpoint

    bands = bands - efermi  # shift all bands to fermi energy 0
    bands = bands[:, start_band:]
    output_bands = orm.ArrayData()
    output_bands.set_array("kpoints", kpoints)
    output_bands.set_array("weights", weights)
    output_bands.set_array("bands", bands)

    if do_smearing:
        # for bands distance convergence
        # and metals in bands measure verification.
        nelectrons = num_electrons
        bands = np.asfortranarray(bands)
        meth = 2  # firmi-dirac smearing

        output_efermi = find_efermi(bands, weights, nelectrons, smearing, meth)

    else:
        # easy to spot the efermi energy only used for non-metals of typical configurations
        # in bands measure.
        homo_energy = get_homo(bands, num_electrons)
        output_efermi = homo_energy

    return {
        "bands": output_bands,
        "efermi": output_efermi,
    }


def calculate_eta_and_max_diff(
    bands_a: orm.ArrayData,
    bands_b: orm.ArrayData,
    efermi_a,
    efermi_b,
    fermi_shift,
    smearing,
):
    """
    calculate the difference of two bands, weight is supported
    """
    from functools import partial

    from scipy.optimize import minimize

    weight_a = bands_a.get_array("weights")
    weight_b = bands_b.get_array("weights")
    weight = weight_a
    assert np.allclose(
        weight_a, weight_b
    ), "Different weight of kpoints of two calculation."

    bands_a = bands_a.get_array("bands")
    bands_b = bands_b.get_array("bands")

    num_bands = min(np.shape(bands_a)[1], np.shape(bands_b)[1])

    assert np.shape(bands_a)[0] == np.shape(bands_b)[0], "have different kpoints"

    # truncate the bands to same size
    bands_a = bands_a[:, : num_bands - 1]
    bands_b = bands_b[:, : num_bands - 1]

    occ_a = fermi_dirac(bands_a, efermi_a + fermi_shift, smearing)
    occ_b = fermi_dirac(bands_b, efermi_b + fermi_shift, smearing)
    occ = np.sqrt(occ_a * occ_b)

    bands_diff = bands_a - bands_b

    def fun_shift(occ, bands_diff, shift):
        # 1/w ~ degeneracy of the kpoints
        nominator = np.multiply(1 / weight[:, None], (occ * (bands_diff + shift) ** 2))
        denominator = np.multiply(1 / weight[:, None], occ)
        return np.sqrt(np.sum(nominator) / np.sum(denominator))

    # Compute eta
    eta_val = partial(fun_shift, occ, bands_diff)
    results = minimize(eta_val, np.array([0.0]), method="Nelder-Mead")
    eta = results.get("fun")
    shift = results.get("x")[0]

    # Compute max_diff:
    # then find from abs_diff the max one of which the occ > 0.5
    # first make occ > 0.5 to be 1 and occ <= 0.5 to be 0, then element-wise the
    # new occ matrix with abs_diff, and find the max diff
    abs_diff = np.abs(bands_diff + shift)
    new_occ = np.heaviside(occ - 0.5, 0.0)
    new_diff = np.multiply(abs_diff, new_occ)
    max_diff = np.amax(new_diff)

    return {
        "eta": eta,
        "shift": shift,
        "max_diff": max_diff,
    }


def get_bands_distance(
    bands_a: orm.BandsData,
    bands_b: orm.BandsData,
    band_parameters_a: orm.Dict,
    band_parameters_b: orm.Dict,
    smearing: float,
    fermi_shift: float,
    do_smearing: bool,
):
    """
    First aligh the number of two bands, e.g tranctrate the overceed nubmer of bands
    """
    num_electrons_a = band_parameters_a["number_of_electrons"]
    num_electrons_b = band_parameters_b["number_of_electrons"]

    assert (
        not num_electrons_a > num_electrons_b
    ), "Need to be less num_electrons result as argument labeled 'a'."

    efermi_a = band_parameters_a["fermi_energy"]
    efermi_b = band_parameters_b["fermi_energy"]

    num_electrons = int(num_electrons_a)
    res = retrieve_bands(bands_a, 0, num_electrons, efermi_a, smearing, do_smearing)
    bands_a = res["bands"]
    efermi_a = res["efermi"]

    start_band = int(num_electrons_b - num_electrons_a) // 2
    res = retrieve_bands(
        bands_b, start_band, num_electrons, efermi_b, smearing, do_smearing
    )
    bands_b = res["bands"]
    efermi_b = res["efermi"]

    # eta_v
    fermi_shift_v = 0.0
    if do_smearing:
        smearing_v = smearing
    else:
        smearing_v = 0

    outputs = calculate_eta_and_max_diff(
        bands_a, bands_b, efermi_a, efermi_b, fermi_shift_v, smearing_v
    )
    eta_v = outputs.get("eta")
    shift_v = outputs.get("shift")
    max_diff_v = outputs.get("max_diff")

    # eta_c
    # if not metal
    smearing_c = smearing
    outputs = calculate_eta_and_max_diff(
        bands_a, bands_b, efermi_a, efermi_b, fermi_shift, smearing_c
    )

    eta_c = outputs.get("eta")
    shift_c = outputs.get("shift")
    max_diff_c = outputs.get("max_diff")

    return {
        "eta_v": eta_v,
        "shift_v": shift_v,
        "max_diff_v": max_diff_v,
        "eta_c": eta_c,
        "shift_c": shift_c,
        "max_diff_c": max_diff_c,
    }



from abc import abstractmethod

from aiida import orm

from aiida_sssp_workflow.workflows import SelfCleanWorkChain


class _BaseEvaluateWorkChain(SelfCleanWorkChain):
    """WorkChain to calculate cohisive energy of input structure"""

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.output('ecutwfc', valid_type=orm.Int, required=True)
        spec.output('ecutrho', valid_type=orm.Int, required=True)

        spec.exit_code(210, 'ERROR_SUB_PROCESS_FAILED_SCF',
                    message='PwBaseWorkChain of pressure scf evaluation failed.')
        # yapf: enable

    @abstractmethod
    def finalize(self):
        """Must in here write ecutwfc and ecutrho to output ports
        They are needed by the convergence workflow to read the cutoff pair
        of evalueation.
        """



from aiida import orm
from aiida.plugins import WorkflowFactory

from aiida_sssp_workflow.utils import update_dict
from aiida_sssp_workflow.workflows.convergence._base import _BaseConvergenceWorkChain

PwBaseWorkflow = WorkflowFactory("quantumespresso.pw.base")


class _CachingConvergenceWorkChain(_BaseConvergenceWorkChain):
    """Convergence caching workflow
    this workflow will only run in verification workflow
    when there are at least two convergence workflows are order to run.
    It also require that the caching machenism of aiida is on.
    The purpose of this workflow is to run a set of common SCF calculations
    with the same input parameters in reference calculation and wavefunction
    cutoff test calculations. In order to save the time and resource for
    the following convergence test."""

    _PROPERTY_NAME = None  # will only use convergence/base protocol
    _EVALUATE_WORKCHAIN = PwBaseWorkflow
    _MEASURE_OUT_PROPERTY = None

    _RUN_WFC_TEST = True
    _RUN_RHO_TEST = False  # will not run charge density cutoff test

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input(
            "clean_workchain",
            valid_type=orm.Bool,
            default=lambda: orm.Bool(False),
            help="If `True`, work directories of all called calculation will be cleaned at the end of execution.",
        )

    def inspect_wfc_convergence_test(self):
        """Override this step to do nothing to parse wavefunction
        cutoff test results but only run it."""
        return None

    def setup_criteria_parameters_from_protocol(self):
        """Override this step to do nothing, since it is not
        used for caching run."""
        return None

    def get_result_metadata(self):
        """No need to actual implemented for caching workchain"""
        return None

    def helper_compare_result_extract_fun(
        self, sample_node, reference_node, **kwargs
    ) -> dict:
        """No need to actual implemented for caching workchain"""
        return None

    def setup_code_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        super().setup_code_parameters_from_protocol()

        protocol = self.ctx.protocol
        self._DEGAUSS = protocol["degauss"]
        self._OCCUPATIONS = protocol["occupations"]
        self._SMEARING = protocol["smearing"]
        self._CONV_THR = protocol["electron_conv_thr"]
        self.ctx.kpoints_distance = self._KDISTANCE = protocol["kpoints_distance"]

        self.ctx.pw_base_parameters = super()._get_pw_base_parameters(
            self._DEGAUSS, self._OCCUPATIONS, self._SMEARING, self._CONV_THR
        )

    def _get_inputs(self, ecutwfc, ecutrho) -> dict:
        """inputs for running a dummy SCF for caching"""
        pw_parameters = update_dict(self.ctx.pw_base_parameters, {})
        pw_parameters["SYSTEM"]["ecutwfc"] = ecutwfc
        pw_parameters["SYSTEM"]["ecutrho"] = ecutrho

        inputs = {
            "metadata": {"call_link_label": "SCF_for_cache"},
            "pw": {
                "structure": self.ctx.structure,
                "code": self.inputs.code,
                "pseudos": self.ctx.pseudos,
                "parameters": orm.Dict(dict=pw_parameters),
                "metadata": {
                    "options": self.ctx.options,
                },
                "parallelization": orm.Dict(dict=self.ctx.parallelization),
            },
            "kpoints_distance": orm.Float(self.ctx.kpoints_distance),
        }

        return inputs
